"""Provide the SpikingJelly single-step with timesteps in input data learningframework module."""

import logging
import sys

import torch
from spikingjelly.activation_based import functional  # type: ignore[import-untyped]

from .SpikingJelly import SpikingJelly

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class SpikingJellyTimeStepsInData(SpikingJelly):
    """SpikingJelly single-step with timesteps in data LearningFramework implementation extending SpikingJelly single-step."""

    class TrainerModule(SpikingJelly.TrainerModule):
        """SpikingJelly single-step with timesteps in data TrainerModule extending SpikingJelly single-step TrainerModule."""

        @override
        def forward(self, x: torch.Tensor) -> torch.Tensor:
            """Forward pass for a Spiking Neural Network model with timesteps in input data in single-step mode.

            First calls SpikingJelly's reset on the model to reset neurons potentials.
            Call :meth:`qualia_plugin_snn.learningmodel.pytorch.SNN.SNN.forward` for each timestep of the input data.
            Finally, average the output of the model over the timesteps.

            :param x: Input data with timestep dimension in [N, C, T, S] or [N, C, T, H, W] order
            :return: Output predictions
            :raise ValueError: when the input data does not have the correct number of dimenions or the timestep dimension does not
                match :attr:`qualia_plugin_snn.learningmodel.pytorch.SNN.SNN.timesteps`
            """
            functional.reset_net(self.model)

            # Switch timestep dim from 3rd to 1st place
            if len(x.shape) == 5: # noqa: PLR2004 2D data, [N, C, T, H, W] → [T, N, C, H, W]
                x = x.permute(2, 0, 1, 3, 4)
            elif len(x.shape) == 4: # noqa: PLR2004 1D data [N, C, T, S] → [T, N, C, S]
                x = x.permute(2, 0, 1, 3)
            else:
                logger.error('Unsupported number of axes in dataset: %s, must be 4 or 5', len(x.shape))
                raise ValueError

            input_timesteps = list(x) # Split timestep dims into multiple arrays
            if len(input_timesteps) != self.model.timesteps:
                logger.error('Model.timesteps differs from timesteps dimension in data: %s != %s',
                             self.model.timesteps,
                             len(input_timesteps))
                raise ValueError

            x_list = [self.model(t) for t in input_timesteps]
            x = torch.stack(x_list).sum(0)

            return x / len(input_timesteps)



"""Pytorch learningmodel templates for Spiking Neural Networks based on SpikingJelly."""

from .QuantizedSCNN import QuantizedSCNN
from .QuantizedSResNet import QuantizedSResNet
from .SCNN import SCNN
from .SResNet import SResNet

__all__ = ['QuantizedSCNN', 'QuantizedSResNet', 'SCNN', 'SResNet']



"""Qualia-Plugin-SNN preprocessing package contains preprocessing modules adapted for or dedicated to Spiking Neural Networks."""

from .Split2TimeSteps import Split2TimeSteps

__all__ = ['Split2TimeSteps']



"""Contains the template for a convolutional spiking neural network."""

from __future__ import annotations

import logging
import math
import sys
from collections import OrderedDict

import numpy as np
import torch
from qualia_core.learningmodel.pytorch import layers1d, layers2d
from qualia_core.typing import TYPE_CHECKING
from torch import nn

from qualia_plugin_snn.learningmodel.pytorch.layers.spikingjelly import layers1d as sjlayers1d
from qualia_plugin_snn.learningmodel.pytorch.layers.spikingjelly import layers2d as sjlayers2d

from .SNN import SNN

if TYPE_CHECKING:
    from types import ModuleType  # noqa: TCH003

    from qualia_core.typing import RecursiveConfigDict

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class SCNN(SNN):
    """Convolutional spiking neural network template.

    Similar to :class:`qualia_core.learningmodel.pytorch.CNN.CNN` but with spiking neuron activation layers (e.g., IF) instead of
    :class:`torch.nn.ReLU`.


    Example TOML configuration for a 2D SVGG16 over 4 timesteps with soft-reset multi-step IF based on the SCNN template:

    .. code-block:: toml

        [[model]]
        name = "VGG16"
        params.filters      = [ 64, 64, 128, 128, 256, 256, 256, 512, 512, 512, 512, 512, 512]
        params.kernel_sizes = [  3,  3,   3,   3,   3,   3,   3,   3,   3,   3,   3,   3,   3]
        params.paddings     = [  1,  1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1]
        params.strides      = [  1,  1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1]
        params.pool_sizes   = [  0,  2,   0,   2,   0,   0,   2,   0,   0,   2,   0,   0,   2]
        params.dropouts     = [  0,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, 0, 0]
        params.fc_units     = [4096, 4096]
        params.batch_norm   = true
        params.timesteps    = 4
        params.dims         = 2
        params.neuron.kind                = 'IFNode'
        params.neuron.params.v_reset      = false # Soft reset
        params.neuron.params.v_threshold  = 1.0
        params.neuron.params.detach_reset = true
        params.neuron.params.step_mode    = 'm' # Multi-step mode, make sure to use SpikingJellyMultiStep learningframework
        params.neuron.params.backend      = 'cupy'
    """

    layers: nn.ModuleDict #: List of sequential layers of the SCNN model

    def __init__(self,  # noqa: PLR0913, PLR0915, PLR0912, C901
                 input_shape: tuple[int, ...],
                 output_shape: tuple[int, ...],

                 filters: list[int],
                 kernel_sizes: list[int],
                 paddings: list[int],
                 strides: list[int],
                 dropouts: float | list[float],
                 pool_sizes: list[int],
                 fc_units: list[int],
                 batch_norm: bool = False,  # noqa: FBT001, FBT002
                 prepool: int | list[int] = 1,
                 postpool: int | list[int] = 1,

                 neuron: RecursiveConfigDict | None = None,
                 timesteps: int = 4,

                 gsp: bool = False,  # noqa: FBT001, FBT002

                 dims: int=1) -> None:
        """Construct :class:`SCNN`.

        :param input_shape: Input shape
        :param output_shape: Output shape
        :param filters: List of ``out_channels`` for each Conv layer, also defines the number of Conv layers
        :param kernel_sizes: List of ``kernel_size`` for each Conv layer, must of the same size as ``filters``
        :param paddings: List of ``padding`` for each Conv layer, must of the same size as ``filters``
        :param strides: List of ``stride`` for each Conv layer, must of the same size as ``filters``
        :param dropouts: List of Dropout layer ``p`` to apply after each Conv or Linear layer, must be of the same size as
                         ``filters`` + ``fc_units``, no layer added if element is 0
        :param pool_sizes: List of MaxPool layer ``kernel_size`` to apply after each Conv layer, must be of the same size as
                           ``filters``, no layer added if element is 0
        :param fc_units: List of :class:`torch.nn.Linear` layer ``out_features`` to add at the end of the network, no layer added
                         if empty
        :param batch_norm: If ``True``, add a BatchNorm layer after each Conv layer, otherwise no layer added
        :param prepool: AvgPool layer ``kernel_size`` to add at the beginning of the network, no layer added if 0
        :param postpool: AvgPool layer ``kernel_size`` to add after all Conv layers, no layer added if 0
        :param neuron: Spiking neuron configuration, see :meth:`qualia_plugin_snn.learningmodel.pytorch.SNN.SNN.__init__`
        :param timesteps: Number of timesteps
        :param gsp: If ``True``, a single GlobalSumPool layer is added instead of Linear layers
        :param dims: Either 1 or 2 for 1D or 2D convolutional network.
        """
        super().__init__(input_shape=input_shape,
                         output_shape=output_shape,
                         timesteps=timesteps,
                         neuron=neuron)

        from spikingjelly.activation_based.layer import Dropout, Flatten, Linear  # type: ignore[import-untyped]

        layers_t: ModuleType
        sjlayers_t: ModuleType

        if dims == 1:
            layers_t = layers1d
            sjlayers_t = sjlayers1d
        elif dims == 2:  # noqa: PLR2004
            layers_t = layers2d
            sjlayers_t = sjlayers2d
        else:
            logger.error('Only dims=1 or dims=2 supported, got: %s', dims)
            raise ValueError

        # Backward compatibility for config not defining dropout as a list
        dropout_list = [dropouts] * (len(filters) + len(fc_units)) if not isinstance(dropouts, list) else dropouts

        layers: OrderedDict[str, nn.Module] = OrderedDict()

        if (math.prod(prepool) if isinstance(prepool, list) else prepool) > 1:
            layers['prepool'] = sjlayers_t.AvgPool(tuple(prepool) if isinstance(prepool, list) else prepool,
                                                   step_mode=self.step_mode)

        layers['conv1'] = sjlayers_t.Conv(in_channels=input_shape[-1],
                                               out_channels=filters[0],
                                               kernel_size=kernel_sizes[0],
                                               padding=paddings[0],
                                               stride=strides[0],
                                               bias=not batch_norm,
                                               step_mode=self.step_mode)

        if batch_norm:
            layers['bn1'] = sjlayers_t.BatchNorm(filters[0], step_mode=self.step_mode)

        layers['neuron1'] = self.create_neuron()

        if dropout_list[0]:
            layers['dropout1'] = Dropout(dropout_list[0], step_mode=self.step_mode)
        if pool_sizes[0]:
            layers['maxpool1'] = sjlayers_t.MaxPool(pool_sizes[0], step_mode=self.step_mode)

        i = 2
        for in_filters, out_filters, kernel, pool_size, padding, stride, dropout in zip(filters, filters[1:], kernel_sizes[1:],
                                                                                        pool_sizes[1:], paddings[1:], strides[1:],
                                                                                        dropout_list[1:]):
            layers[f'conv{i}'] = sjlayers_t.Conv(in_channels=in_filters,
                                                      out_channels=out_filters,
                                                      kernel_size=kernel,
                                                      padding=padding,
                                                      stride=stride,
                                                      bias=not batch_norm,
                                                      step_mode=self.step_mode)

            if batch_norm:
                layers[f'bn{i}'] = sjlayers_t.BatchNorm(out_filters, step_mode=self.step_mode)

            layers[f'neuron{i}'] = self.create_neuron()

            if dropout:
                layers[f'dropout{i}'] = Dropout(dropout, step_mode=self.step_mode)
            if pool_size:
                layers[f'maxpool{i}'] = sjlayers_t.MaxPool(pool_size, step_mode=self.step_mode)

            i += 1

        if (math.prod(postpool) if  isinstance(postpool, list) else postpool) > 1:
            layers['postpool'] = sjlayers_t.AvgPool(tuple(postpool) if isinstance(postpool, list) else postpool,
                                                    step_mode=self.step_mode)

        if gsp:
            layers['gsp'] = layers_t.GlobalSumPool()
        else:
            layers['flatten'] = Flatten(step_mode=self.step_mode)

            in_features = np.array(input_shape[:-1]) // np.array(prepool)
            for _, kernel, pool, padding, stride in zip(filters, kernel_sizes, pool_sizes, paddings, strides):
                in_features += np.array(padding) * 2
                in_features -= (kernel - 1)
                in_features = np.ceil(in_features / stride).astype(int)
                if pool:
                    in_features = in_features // pool
            in_features = in_features // np.array(postpool)
            in_features = in_features.prod()
            in_features *= filters[-1]

            j = 1
            for in_units, out_units, dropout in zip((in_features, *fc_units), fc_units, dropout_list[len(filters):]):
                layers[f'fc{j}'] = Linear(in_units, out_units, step_mode=self.step_mode)
                layers[f'neuron{i}'] = self.create_neuron()
                if dropout:
                    layers[f'dropout{i}'] = Dropout(dropout, step_mode=self.step_mode)
                i += 1
                j += 1

            layers[f'fc{j}'] = Linear(fc_units[-1] if len(fc_units) > 0 else in_features,
                                           output_shape[0],
                                           step_mode=self.step_mode)

        self.layers = nn.ModuleDict(layers)

    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002
        """Forward calls each of the SCNN :attr:`layers` sequentially.

        :param input: Input tensor
        :return: Output tensor
        """
        x = input
        for layer in self.layers:
            x = self.layers[layer](x)

        return x



"""Provide the Qualia-CodeGen Linux deployer class with support for Spiking Neural Networks."""

from __future__ import annotations

from importlib.resources import files

from qualia_core.deployment.qualia_codegen.Linux import Linux as LinuxQualiaCore
from qualia_core.utils.path import resources_to_path


class Linux(LinuxQualiaCore):
    """Qualia-CodeGen Linux deployer using example from qualia_codegen-plugin-snn for SNN support."""

    def __init__(self,
                 cxxflags: list[str] | None = None,
                 modeldir: str | None = None,
                 projectdir: str | None = None,
                 outdir: str | None = None) -> None:
        """Construct :class:`qualia_plugin_snn.deployment.qualia_codegen.Linux.Linux`.

        :param cxxflags: Override default compiler flags, see :meth:`qualia_core.deployment.qualia_codegen.Linux.Linux.__init__`
        :param modeldir: Path to model C code directory, default: ``out/qualia_codegen``
        :param projectdir: Path to Qualia-CodeGen-Plugin-SNN Linux project dir, default:
            ``<qualia_codegen_plugin_snn.examples>/Linux``
        :param outdir: Path to build products directory, default: ``out/deploy/Linux``
        """
        super().__init__(cxxflags=cxxflags,
                         modeldir=modeldir,
                         projectdir=projectdir if projectdir is not None else
                            resources_to_path(files('qualia_codegen_plugin_snn.examples')/'Linux'),
                         outdir=outdir)



"""Aliases for 2D SpikingJelly and quantized SpikingJelly layers."""

from spikingjelly.activation_based.layer import AvgPool2d as AvgPool  # type: ignore[import-untyped]
from spikingjelly.activation_based.layer import BatchNorm2d as BatchNorm
from spikingjelly.activation_based.layer import Conv2d as Conv
from spikingjelly.activation_based.layer import MaxPool2d as MaxPool

from .quantized_layers2d import QuantizedBatchNorm2d as QuantizedBatchNorm
from .quantized_layers2d import QuantizedConv2d as QuantizedConv
from .quantized_layers2d import QuantizedMaxPool2d as QuantizedMaxPool

__all__ = ['Conv',
           'BatchNorm',
           'AvgPool',
           'MaxPool',
           'QuantizedBatchNorm',
           'QuantizedConv',
           'QuantizedMaxPool']



"""Contain implementation of quantized 1D layers with support for SpikingJelly ``step_mode``."""
from __future__ import annotations

import logging
import sys
from typing import cast

from qualia_core.learningmodel.pytorch import quantized_layers1d
from qualia_core.typing import TYPE_CHECKING
from spikingjelly.activation_based import functional  # type: ignore[import-untyped]
from spikingjelly.activation_based.base import StepModule  # type: ignore[import-untyped]

# We are inside a TYPE_CHECKING block but our custom TYPE_CHECKING constant triggers TCH001-TCH003 so ignore them
if TYPE_CHECKING:
    import torch  # noqa: TCH002
    from qualia_core.learningmodel.pytorch.Quantizer import QuantizationConfig  # noqa: TCH002
    from torch import nn  # noqa: TCH002

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class QuantizedConv1d(quantized_layers1d.QuantizedConv1d,
                      StepModule):  # type: ignore[misc]
    """Add SpikingJelly's ``step_mode`` support to Qualia's quantized Conv1d layer."""

    def __init__(self,  # noqa: PLR0913
                 in_channels: int,
                 out_channels: int,
                 quant_params: QuantizationConfig,
                 kernel_size: int = 3,
                 stride: int = 1,
                 padding: int = 0,
                 dilation: int = 1,
                 groups: int = 1,
                 bias: bool = True,  # noqa: FBT001, FBT002
                 activation: nn.Module | None = None,
                 step_mode: str = 's') -> None:
        """Construct :class:`QuantizedConv1d`.

        :param in_channels: Number of input channels
        :param out_channels: Number of output channels, i.e., number of filters
        :param kernel_size: Size of the kernel
        :param stride: Stride of the convolution
        :param padding: Padding added to both sides of the input
        :param dilation: Spacing between kernel elements
        :param groups: Number of blocked connections from input channels to output channels
        :param bias: If ``True``, adds a learnable bias to the output
        :param quant_params: Dict containing quantization parameters, see
                             :class:`qualia_core.learningmodel.pytorch.Quantizer.Quantizer`
        :param activation: Activation layer to fuse for quantization purposes, ``None`` if unfused or no activation
        :param step_mode: SpikingJelly's ``step_mode``, either ``'s'`` or ``'m'``, see
                          :class:`spikingjelly.activation_based.layer.Linear`
        """
        super().__init__(in_channels=in_channels,
                         out_channels=out_channels,
                         kernel_size=kernel_size,
                         stride=stride,
                         padding=padding,
                         dilation=dilation,
                         groups=groups,
                         bias=bias,
                         quant_params=quant_params,
                         activation=activation)
        self.step_mode = step_mode

    @override
    def extra_repr(self) -> str:
        """Add ``step_mode`` to the ``__repr__`` method.

        :return: String representation of :class:`torch.nn.Conv1d` with ``step_mode``.
        """
        return cast(str, super().extra_repr()) + f', step_mode={self.step_mode}'  # type: ignore[no-untyped-call]

    # From spikingjelly.activation_based.layer.Conv1d
    # Copied to call super().forward(x) of quantized_layers1d.Conv1d instead of torch.nn.Conv1d
    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002
        """Forward :class:`qualia_core.learningmodel.pytorch.quantized_layers1d.QuantizedConv1d` with ``step_mode`` support.

        :param input: Input tensor
        :return: Output tensor
        """
        x = input
        if self.step_mode == 's':
            x = super().forward(x)

        elif self.step_mode == 'm':
            if x.dim() != 4:  # noqa: PLR2004
                logger.error('Expected input with shape [T, N, C, L], but got input with shape %s', x.shape)
                raise ValueError
            x = functional.seq_to_ann_forward(x, super().forward)

        return x

class QuantizedMaxPool1d(quantized_layers1d.QuantizedMaxPool1d,
                         StepModule):  # type: ignore[misc]
    """Add SpikingJelly's ``step_mode`` support to Qualia's quantized MaxPool1d layer."""

    def __init__(self,  # noqa: PLR0913
                 kernel_size: int,
                 quant_params: QuantizationConfig,
                 stride: int | None = None,
                 padding: int = 0,
                 activation: nn.Module | None = None,
                 step_mode: str = 's') -> None:
        """Construct :class:`QuantizedMaxPool1d`.

        :param kernel_size: Size of the sliding window
        :param stride: Stride of the sliding window
        :param padding:  Implicit negative infinity padding to be added on both sides
        :param quant_params: Dict containing quantization parameters, see
                             :class:`qualia_core.learningmodel.pytorch.Quantizer.Quantizer`
        :param activation: Activation layer to fuse for quantization purposes, ``None`` if unfused or no activation
        :param step_mode: SpikingJelly's ``step_mode``, either ``'s'`` or ``'m'``, see
                          :class:`spikingjelly.activation_based.layer.Linear`
        """
        super().__init__(kernel_size=kernel_size,
                         stride=stride,
                         padding=padding,
                         quant_params=quant_params,
                         activation=activation)
        self.step_mode = step_mode

    @override
    def extra_repr(self) -> str:
        """Add ``step_mode`` to the ``__repr__`` method.

        :return: String representation of :class:`torch.nn.MaxPool1d` with ``step_mode``.
        """
        return super().extra_repr() + f', step_mode={self.step_mode}'

    # From spikingjelly.activation_based.layer.MaxPool1d
    # Copied to call super().forward(x) of quantized_layers1d.MaxPool1d instead of torch.nn.MaxPool1d
    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002
        """Forward :class:`qualia_core.learningmodel.pytorch.quantized_layers1d.QuantizedMaxPool1d` with ``step_mode`` support.

        :param input: Input tensor
        :return: Output tensor
        """
        x = input
        if self.step_mode == 's':
            x = super().forward(x)

        elif self.step_mode == 'm':
            if x.dim() != 4:  # noqa: PLR2004
                logger.error('Expected input with shape [T, N, C, L], but got input with shape %s', x.shape)
                raise ValueError
            x = functional.seq_to_ann_forward(x, super().forward)

        return x

class QuantizedBatchNorm1d(quantized_layers1d.QuantizedBatchNorm1d,
                           StepModule):  # type: ignore[misc]
    """Add SpikingJelly's ``step_mode`` support to Qualia's quantized BatchNorm1d layer."""

    def __init__(self,  # noqa: PLR0913
                 num_features: int,
                 quant_params: QuantizationConfig,
                 eps: float = 1e-05,
                 momentum: float = 0.1,
                 affine: bool = True,  # noqa: FBT001, FBT002
                 track_running_stats: bool = True,  # noqa: FBT001, FBT002
                 device: torch.device | None = None,
                 dtype: torch.dtype | None = None,
                 activation: torch.nn.Module | None = None,
                 step_mode: str = 's') -> None:
        """Construct :class:`QuantizedBatchNorm1d`.

        :param num_features: Number of features or channels of the input
        :param eps: Value added to the denominator for numerical stability
        :param momentum: value used for the running_mean and running_var computation
        :param affine: If ``True``, add learnable affine parameters
        :param track_running_stats: If ``True``, track the running mean and variance, otherwise always uses batch statistics
        :param device: Optional device to perform the computation on
        :param dtype: Optional data type
        :param quant_params: Dict containing quantization parameters, see
                             :class:`qualia_core.learningmodel.pytorch.Quantizer.Quantizer`
        :param activation: Activation layer to fuse for quantization purposes, ``None`` if unfused or no activation
        :param step_mode: SpikingJelly's ``step_mode``, either ``'s'`` or ``'m'``, see
                          :class:`spikingjelly.activation_based.layer.Linear`
        """
        super().__init__(num_features=num_features,
                         eps=eps,
                         momentum=momentum,
                         affine=affine,
                         track_running_stats=track_running_stats,
                         device=device,
                         dtype=dtype,
                         quant_params=quant_params,
                         activation=activation)
        self.step_mode = step_mode

    @override
    def extra_repr(self) -> str:
        """Add ``step_mode`` to the ``__repr__`` method.

        :return: String representation of :class:`torch.nn.BatchNorm1d` with ``step_mode``.
        """
        return super().extra_repr() + f', step_mode={self.step_mode}'

    # From spikingjelly.activation_based.layer.MaxPool1d
    # Copied to call super().forward(x) of quantized_layers1d.MaxPool1d instead of torch.nn.MaxPool1d
    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002
        """Forward :class:`qualia_core.learningmodel.pytorch.quantized_layers1d.QuantizedBatchNorm1d` with ``step_mode`` support.

        :param input: Input tensor
        :return: Output tensor
        """
        x = input
        if self.step_mode == 's':
            x = super().forward(x)

        elif self.step_mode == 'm':
            if x.dim() != 4:  # noqa: PLR2004
                logger.error('Expected input with shape [T, N, C, L], but got input with shape %', x.shape)
                raise ValueError
            x = functional.seq_to_ann_forward(x, super().forward)

        return x



"""Contains implementation for custom or quantized layers for spiking neural networks."""

from .CustomNode import ATIF, IFSRL

custom_nodes = (ATIF, IFSRL)
__all__ = ['ATIF', 'IFSRL']

layers = custom_nodes



"""Qualia-Plugin-SNN plugin.

When Qualia's configuration file contains:

.. code-block:: toml

    [bench]
    plugin = ['qualia_plugin_snn']

The following subpackages are imported:

* :mod:`qualia_plugin_snn.preprocessing`
* :mod:`qualia_plugin_snn.learningframework`
* :mod:`qualia_plugin_snn.postprocessing`

Subpackage :mod:`qualia_plugin_snn.deployment` contains deployers for Qualia-CodeGen referenced in
:attr:`qualia_plugin_snn.postprocessing.QualiaCodeGen.QualiaCodeGen.deployers`.

Subpackage :mod:`qualia_plugin_snn.learningframework` contains the LearningFramework implementation for SpikingJelly.

Subpackage :mod:`qualia_plugin_snn.learningmodel` contains the spiking neural network templates referenced in
:attr:`qualia_plugin_snn.learningframework.SpikingJelly.SpikingJelly.learningmodels`, made available when the SpikingJelly
learningframework is used.

Subpackage :mod:`qualia_plugin_snn.preprocessing` contains preprocessing modules adapted for or dedicated to Spiking Neural
Networks.

Subpackage :mod:`qualia_plugin_snn.postprocessing` contains postprocessing modules adapted for or dedicated to
Spiking Neural Networks, in particular the Qualia-CodeGen interface relying on :mod:`qualia_codegen_plugin_snn`,
part of Qualia-CodeGen-Plugin-SNN.
"""



"""Aliases for 1D SpikingJelly and quantized SpikingJelly layers."""

from spikingjelly.activation_based.layer import AvgPool1d as AvgPool  # type: ignore[import-untyped]
from spikingjelly.activation_based.layer import BatchNorm1d as BatchNorm
from spikingjelly.activation_based.layer import Conv1d as Conv
from spikingjelly.activation_based.layer import MaxPool1d as MaxPool

from .quantized_layers1d import QuantizedBatchNorm1d as QuantizedBatchNorm
from .quantized_layers1d import QuantizedConv1d as QuantizedConv
from .quantized_layers1d import QuantizedMaxPool1d as QuantizedMaxPool

__all__ = ['Conv',
           'BatchNorm',
           'AvgPool',
           'MaxPool',
           'QuantizedBatchNorm',
           'QuantizedConv',
           'QuantizedMaxPool']



"""Provide the model postprocessing class to generate C code for Spiking Neural Networks using Qualia-CodeGen-Plugin-SNN."""

from __future__ import annotations

import sys
from pathlib import Path
from typing import Any, Callable

import qualia_core.postprocessing
from qualia_core.typing import TYPE_CHECKING

import qualia_plugin_snn.deployment.qualia_codegen

# We are inside a TYPE_CHECKING block but our custom TYPE_CHECKING constant triggers TCH001-TCH003 so ignore them
if TYPE_CHECKING:
    from types import ModuleType  # noqa: TCH003

    from qualia_codegen_core.graph import ModelGraph  # noqa: TCH002
    from torch import nn  # noqa: TCH002

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class QualiaCodeGen(qualia_core.postprocessing.QualiaCodeGen):
    """Qualia-CodeGen converter calling Qualia-CodeGen-Plugin-SNN to handle Spiking Neural Network layers."""

    deployers: ModuleType = qualia_plugin_snn.deployment.qualia_codegen
    """:mod:`qualia_plugin_snn.deployment.qualia_codegen` default deployers.

    Includes :class:`qualia_plugin_spleat.deployment.qualia_codegen.Linux.Linux`.

    :meta hide-value:
    """

    @override
    def convert_model_to_modelgraph(self, model: nn.Module) -> ModelGraph | None:
        """Convert PyTorch model to a Qualia-CodeGen ModelGraph graph representation.

        Uses the :class:`qualia_codegen_plugin_snn.graph.TorchModelGraph.TorchModelGraph` from Qualia-CodeGen-Plugin-SNN in order
        to support Spiking Neural Networks.
        The following layers are passed as ``custom_layers``:

        * :class:`spikingjelly.activation_based.layer.Conv1d`, handled like a :class:`torch.nn.Conv1d`
        * :class:`spikingjelly.activation_based.layer.Conv2d`, handled like a :class:`torch.nn.Conv2d`
        * :class:`spikingjelly.activation_based.layer.Flatten`, handled like a :class:`torch.nn.Flatten`
        * :class:`spikingjelly.activation_based.layer.Linear`, handled like a :class:`torch.nn.Linear`
        * :class:`qualia_core.learningmodel.pytorch.layers.Add.Add`, mapped to a
          :class:`qualia_codegen_core.graph.layers.TAddLayer.TAddLayer`
        * :class:`qualia_core.learningmodel.pytorch.layers.GlobalSumPool1d.GlobalSumPool1d`, mapped to a
          :class:`qualia_codegen_core.graph.layers.TSumLayer.TSumLayer`
        * :class:`qualia_core.learningmodel.pytorch.layers.GlobalSumPool2d.GlobalSumPool2d`, mapped to a
          :class:`qualia_codegen_core.graph.layers.TSumLayer.TSumLayer`

        SpikingJelly ``step_mode`` is forced to ``'s'`` for single-step operation to simplify visit of the graph.

        :param model: PyTorch model
        :return: Qualia-CodeGen ModelGraph or None in case of error
        """
        import spikingjelly.activation_based.functional as sjf  # type: ignore[import-untyped]
        import spikingjelly.activation_based.layer as sjl  # type: ignore[import-untyped]
        from qualia_codegen_core.graph.layers import TAddLayer, TBaseLayer, TSumLayer
        from qualia_codegen_plugin_snn.graph import TorchModelGraph
        from qualia_core.learningmodel.pytorch.layers import Add, GlobalSumPool1d, GlobalSumPool2d
        from torch import nn

        custom_layers: dict[type[nn.Module], Callable[[nn.Module, TBaseLayer], tuple[type[TBaseLayer], list[Any]]]]= {
                sjl.Conv1d: TorchModelGraph.MODULE_MAPPING[nn.Conv1d],
                sjl.Conv2d: TorchModelGraph.MODULE_MAPPING[nn.Conv2d],
                sjl.Flatten: TorchModelGraph.MODULE_MAPPING[nn.Flatten],
                sjl.Linear: TorchModelGraph.MODULE_MAPPING[nn.Linear],
                Add: lambda *_: (TAddLayer, []),
                GlobalSumPool1d: lambda *_: (TSumLayer, [(-1,)]),
                GlobalSumPool2d: lambda *_: (TSumLayer, [(-2, -1)]),
                }
        sjf.set_step_mode(model, step_mode='s')

        return TorchModelGraph(model).convert(custom_layers=custom_layers)

    @override
    def convert_modelgraph_to_c(self, modelgraph: ModelGraph, output_path: Path) -> str | None:
        """Generate C code for the given ModelGraph using Qualia-CodeGen.

        Uses the :class:`qualia_codegen_plugin_snn.Converter.Converter` from Qualia-CodeGen-Plugin-SNN in order to support Spiking
        Neural Networks.

        :param modelgraph: The ModelGraph object from :class:`qualia_codegen_plugin_snn.graph.TorchModelGraph.TorchModelGraph`
            after conversion with :meth:`convert_model_to_modelgraph`
        :param output_path: Generated C code output path
        :return: String containing the single-file C code
        """
        from qualia_codegen_plugin_snn import Converter
        converter = Converter(output_path=output_path)
        return converter.convert_model(modelgraph)



"""Contains the template for a quantized residual spiking neural network."""

from __future__ import annotations

import logging
import sys
from typing import Protocol, cast

import numpy as np
import torch
from qualia_core.learningmodel.pytorch.layers import QuantizedAdd
from qualia_core.learningmodel.pytorch.layers.quantized_layers import QuantizedIdentity
from qualia_core.typing import TYPE_CHECKING
from torch import nn

from qualia_plugin_snn.learningmodel.pytorch.layers.spikingjelly import layers1d as sjlayers1d
from qualia_plugin_snn.learningmodel.pytorch.layers.spikingjelly import layers2d as sjlayers2d

from .SNN import SNN

if TYPE_CHECKING:
    from types import ModuleType  # noqa: TCH003

    from qualia_core.learningmodel.pytorch.Quantizer import QuantizationConfig  # noqa: TCH002
    from qualia_core.typing import RecursiveConfigDict

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class CreateNeuron(Protocol):
    """Signature for create_neuron from :meth:`qualia_plugin_snn.learningmodel.pytorch.SNN.create_neuron`.

    Used to pass neuron builder to BasicBlock.
    """

    def __call__(self, quant_params: QuantizationConfig) -> nn.Module:
        """Instanciate a spiking neuron.

        :param quant_params: Optional quantization configuration dict in case of quantized network, see
                             :class:`qualia_core.learningmodel.pytorch.Quantizer.Quantizer`
        :return: A spiking neuron instance
        """
        ...

class QuantizedBasicBlockBuilder(Protocol):
    """Signature for basicblockbuilder.

    Used to bind hyperparameters constant across all the ResNet blocks.
    """

    def __call__(self,  # noqa: PLR0913
                 in_planes: int,
                 planes: int,
                 kernel_size: int,
                 stride: int,
                 padding: int) -> QuantizedBasicBlock:
        """Build a :class:`BasicBlock`.

        :param in_planes: Number of input channels
        :param planes: Number of filters (i.e., output channels) in the main branch Conv layers
        :param kernel_size: ``kernel_size`` for the main branch Conv layers
        :param stride: ``kernel_size`` for the MaxPool layers, no MaxPool layer added if `1`
        :param padding: Padding for the main branch Conv layers
        :return: A :class:`BasicBlock`
        """
        ...

class QuantizedBasicBlock(nn.Module):
    r"""A single quantized ResNetv1 block.

    Should have topology identical to :class:`qualia_plugin_snn.learningmodel.pytorch.SResNet.BasicBlock` but with layers replaced
    with their quantized equivalent.

    Structure is:

    .. code-block::

                         |
                     /        \
                 |                 |
           QuantizedConv           |
                 |                 |
        QuantizedBatchNorm         |
                 |                 |
         QuantizedMaxPool   QuantizedConv
                 |                 |
            QuantizedIF  QuantizedBatchNorm
                 |                 |
           QuantizedConv  QuantizedMaxPool
                 |                 |
        QuantizedBatchNorm   QuantizedIF
                 |                 |
            QuantizedIF            |
                 |                 |
                     \       /
                         |
                    QuantizedAdd

    Main (left) branch QuantizedConv use ``kernel_size=kernel_size``, while residual (right) branch QuantizedConv
    use ``kernel_size=1``.

    QuantizedBatchNorm layers will be absent if ``batch_norm == False``

    QuantizedMaxPool layer will be absent if ``stride == 1``.

    Residual (right) branch QuantizedConv layer will be asbent if ``in_planes==planes``, except if
    ``force_projection_with_stride==True`` and ``stride != 1``.
    """

    expansion: int = 1 #: Unused

    def __init__(self,  # noqa: PLR0913
                 sjlayers_t: ModuleType,
                 in_planes: int,
                 planes: int,
                 kernel_size: int,
                 stride: int,
                 padding: int,
                 batch_norm: bool,  # noqa: FBT001
                 bn_momentum: float,
                 force_projection_with_stride: bool,  # noqa: FBT001
                 create_neuron: CreateNeuron,
                 step_mode: str,
                 quant_params: QuantizationConfig) -> None:
        """Construct :class:`QuantizedBasicBlock`.

        :param sjlayers_t: Module containing the aliased quantized layers to use (1D or 2D)
        :param in_planes: Number of input channels
        :param planes: Number of filters (i.e., output channels) in the main branch QuantizedConv layers
        :param kernel_size: ``kernel_size`` for the main branch QuantizedConv layers
        :param stride: ``kernel_size`` for the QuantizedMaxPool layers, no QuantizedMaxPool layer added if `1`
        :param padding: Padding for the main branch QuantizedConv layers
        :param batch_norm: If ``True``, add BatchNorm layer after each QuantizedConv layer
        :param bn_momentum: QuantizedBatchNorm layer ``momentum``
        :param force_projection_with_stride: If ``True``, residual QuantizedConv layer is kept when ``stride != 1`` even if
                                             ``in_planes == planes``
        :param create_neuron: :meth:`qualia_plugin_snn.learningmodel.pytorch.SNN.SNN.create_neuron` method to instantiate a spiking
                              neuron
        :param step_mode: SpikingJelly ``step_mode`` from :attr:`qualia_plugin_snn.learningmodel.pytorch.SNN.SNN.step_mode`
        :param quant_params: Quantization configuration dict, see :class:`qualia_core.learningmodel.pytorch.Quantizer.Quantizer`
        """
        super().__init__()
        self.expansion = 1
        self.batch_norm = batch_norm
        self.force_projection_with_stride = force_projection_with_stride
        self.in_planes = in_planes
        self.planes = planes
        self.stride = stride

        self.conv1 = sjlayers_t.QuantizedConv(in_planes,
                                              planes,
                                              kernel_size=kernel_size,
                                              stride=1,
                                              padding=padding,
                                              bias=not batch_norm,
                                              step_mode=step_mode,
                                              quant_params=quant_params,
                                              activation=None)
        if batch_norm:
            self.bn1 = sjlayers_t.QuantizedBatchNorm(planes, momentum=bn_momentum, step_mode=step_mode, quant_params=quant_params)

        if self.stride != 1:
            self.pool1 = sjlayers_t.QuantizedMaxPool(stride, step_mode=step_mode, quant_params=quant_params, activation=None)

        self.neuron1 = create_neuron(quant_params=quant_params)

        self.conv2 = sjlayers_t.QuantizedConv(planes,
                                              planes,
                                              kernel_size=kernel_size,
                                              stride=1,
                                              padding=padding,
                                              bias=not batch_norm,
                                              step_mode=step_mode,
                                              quant_params=quant_params)

        if batch_norm:
            self.bn2 = sjlayers_t.QuantizedBatchNorm(planes, momentum=bn_momentum, step_mode=step_mode, quant_params=quant_params)

        self.neuron2 = create_neuron(quant_params=quant_params)

        if self.stride != 1:
            self.smax = sjlayers_t.QuantizedMaxPool(stride, step_mode=step_mode, quant_params=quant_params)
        if self.in_planes != self.expansion*self.planes or force_projection_with_stride and self.stride != 1:
            self.sconv = sjlayers_t.QuantizedConv(in_planes,
                                                  self.expansion*planes,
                                                  kernel_size=1,
                                                  stride=1,
                                                  bias=not batch_norm,
                                                  step_mode=step_mode,
                                                  quant_params=quant_params)
            self.neuronr = create_neuron(quant_params=quant_params)

            if batch_norm:
                self.sbn = sjlayers_t.QuantizedBatchNorm(self.expansion*planes,
                                                         momentum=bn_momentum,
                                                         step_mode=step_mode,
                                                         quant_params=quant_params)
        self.add = QuantizedAdd(quant_params=quant_params , activation=None)

    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002
        """Forward of quantized ResNet block.

        :param input: Input tensor
        :return: Output tensor
        """
        x = input

        out = self.conv1(x)

        if self.batch_norm:
            out = self.bn1(out)
        if self.stride != 1:
            out = self.pool1(out)
        out = self.neuron1(out)

        out = self.conv2(out)
        if self.batch_norm:
            out = self.bn2(out)
        out = self.neuron2(out)

        # shortcut
        tmp = x

        if self.in_planes != self.expansion*self.planes or self.force_projection_with_stride and self.stride != 1:
            tmp = self.sconv(tmp)

            if self.batch_norm:
                tmp = self.sbn(tmp)

        if self.stride != 1:
            tmp = self.smax(tmp)

        if self.in_planes != self.expansion*self.planes or self.force_projection_with_stride and self.stride != 1:
            tmp = self.neuronr(tmp)

        return self.add(out, tmp)

class QuantizedSResNet(SNN):
    """Quantized residual spiking neural network template.

    Should have topology identical to :class:`qualia_plugin_snn.learningmodel.pytorch.SResNet.SResNet` but with layers replaced
    with their quantized equivalent.
    """

    def __init__(self,  # noqa: PLR0913, C901
                 input_shape: tuple[int, ...],
                 output_shape: tuple[int, ...],
                 filters: list[int],
                 kernel_sizes: list[int],

                 num_blocks: list[int],
                 strides: list[int],
                 paddings: list[int],

                 quant_params: QuantizationConfig,

                 prepool: int = 1,
                 postpool: str = 'max',
                 batch_norm: bool = False,  # noqa: FBT001, FBT002
                 bn_momentum: float = 0.1,
                 force_projection_with_stride: bool = True,  # noqa: FBT001, FBT002

                 neuron: RecursiveConfigDict | None = None,
                 timesteps: int = 2,

                 dims: int = 1,
                 basicblockbuilder: QuantizedBasicBlockBuilder | None = None) -> None:
        """Construct :class:`QuantizedSResNet`.

        Structure is:

        .. code-block::

              QuantizedInput
                     |
             QuantizedAvgPool
                     |
               QuantizedConv
                     |
            QuantizedBatchNorm
                     |
                QuantizedIF
                     |
            QuantizedBasicBlock
                     |
                     …
                     |
            QuantizedBasicBlock
                     |
            QuantizedGlobalPool
                     |
                  Flatten
                     |
              QuantizedLinear

        :param input_shape: Input shape
        :param output_shape: Output shape
        :param filters: List of ``out_channels`` for QuantizedConv layers inside each :class:`QuantizedBasicBlock` group, must be
                        of the same size as ``num_blocks``, first element is for the first QuantizedConv layer at the beginning
                        of the network
        :param kernel_sizes: List of ``kernel_size`` for QuantizedConv layers inside each :class:`QuantizedBasicBlock` group, must
                             of the same size as ``num_blocks``, first element is for the first QuantizedConv layer at the
                             beginning of the network
        :param num_blocks: List of number of :class:`QuantizedBasicBlock` in each group, also defines the number
                           of :class:`QuantizedBasicBlock` groups inside the network
        :param strides: List of ``kernel_size`` for QuantizedMaxPool layers inside each :class:`QuantizedBasicBlock` group, must
                        of the same size as ``num_blocks``, ``stride`` is applied only to the first :class:`QuantizedBasicBlock`
                        of the group, next :class:`QuantizedBasicBlock` in the group use a ``stride`` of ``1``, first element is
                        the stride of the first QuantizedConv layer at the beginning of the network
        :param paddings: List of ``padding`` for QuantizedConv layer inside each :class:`QuantizedBasicBlock` group, must of the
                         same size as ``num_blocks``, first element is for the first QuantizedConv layer at the beginning
                         of the network
        :param prepool: QuantizedAvgPool layer ``kernel_size`` to add at the beginning of the network, no layer added if 0
        :param postpool: Quantized global pooling layer type after all :class:`QuantizedBasicBlock`, either `max` for
                         QuantizedMaxPool or `avg` for QuantizedAvgPool
        :param batch_norm: If ``True``, add a QuantizedBatchNorm layer after each QuantizedConv layer, otherwise no layer added
        :param bn_momentum: QuantizedBatchNorm ``momentum``
        :param force_projection_with_stride: If ``True``, residual QuantizedConv layer is kept when ``stride != 1`` even if
                                             ``in_planes == planes`` inside a :class:`QuantizedBasicBlock`
        :param neuron: Spiking neuron configuration, see :meth:`qualia_plugin_snn.learningmodel.pytorch.SNN.SNN.__init__`
        :param timesteps: Number of timesteps
        :param dims: Either 1 or 2 for 1D or 2D convolutional network.
        :param basicblockbuilder: Optional function with :meth:`QuantizedBasicBlockBuilder.__call__` signature to build a basic
                                  block after binding constants common across all basic blocks
        """
        # Prepend Quantized to the neuron kind to instantiate quantized spiking neurons
        if neuron is not None and 'kind' in neuron and isinstance(neuron['kind'], str):
            neuron['kind'] = 'Quantized' + neuron['kind']

        super().__init__(input_shape=input_shape,
                         output_shape=output_shape,
                         timesteps=timesteps,
                         neuron=neuron)

        from spikingjelly.activation_based.layer import Flatten  # type: ignore[import-untyped]

        from qualia_plugin_snn.learningmodel.pytorch.layers.spikingjelly.quantized_layers import QuantizedLinear


        sjlayers_t: ModuleType
        if dims == 1:
            sjlayers_t = sjlayers1d
        elif dims == 2:  # noqa: PLR2004
            sjlayers_t = sjlayers2d
        else:
            logger.error('Only dims=1 or dims=2 supported, got: %s', dims)
            raise ValueError

        if basicblockbuilder is None:
            def builder(in_planes: int,
                        planes: int,
                        kernel_size: int,
                        stride: int,
                        padding: int) -> QuantizedBasicBlock:
                return QuantizedBasicBlock(
                        sjlayers_t=sjlayers_t,
                        in_planes=in_planes,
                        planes=planes,
                        kernel_size=kernel_size,
                        stride=stride,
                        padding=padding,
                        batch_norm=batch_norm,
                        bn_momentum=bn_momentum,
                        force_projection_with_stride=force_projection_with_stride,
                        create_neuron=self.create_neuron,
                        step_mode=self.step_mode,
                        quant_params=quant_params)
            basicblockbuilder = builder

        self.in_planes: int = filters[0]
        self.batch_norm = batch_norm
        self.num_blocks = num_blocks

        self.identity1 = QuantizedIdentity(quant_params=quant_params)

        if prepool > 1:
            self.prepool = sjlayers_t.QuantizedAvgPool(prepool, step_mode=self.step_mode, quant_params=quant_params)

        self.conv1 = sjlayers_t.QuantizedConv(input_shape[-1],
                                              filters[0],
                                              kernel_size=kernel_sizes[0],
                                              stride=strides[0],
                                              padding=paddings[0],
                                              bias=not batch_norm,
                                              step_mode=self.step_mode,
                                              quant_params=quant_params)
        if self.batch_norm:
            self.bn1 = sjlayers_t.QuantizedBatchNorm(self.in_planes,
                                                     momentum=bn_momentum,
                                                     step_mode=self.step_mode,
                                                     quant_params=quant_params)

        self.neuron1: nn.Module = self.create_neuron(quant_params=quant_params)

        blocks: list[nn.Sequential] = []
        for planes, kernel_size, stride, padding, num_block in zip(filters[1:],
                                                                   kernel_sizes[1:],
                                                                   strides[1:],
                                                                   paddings[1:],
                                                                   num_blocks):
            blocks.append(self._make_layer(basicblockbuilder, num_block, planes, kernel_size, stride, padding))
        self.layers = nn.ModuleList(blocks)

        # GlobalMaxPool kernel_size computation
        self._fm_dims = np.array(input_shape[:-1]) // np.array(prepool)
        for _, kernel, stride, padding in zip(filters, kernel_sizes, strides, paddings):
            self._fm_dims += np.array(padding) * 2
            self._fm_dims -= (kernel - 1)
            self._fm_dims = np.floor(self._fm_dims / stride).astype(int)


        if postpool == 'avg':
            self.postpool = sjlayers_t.QuantizedAvgPool(tuple(self._fm_dims), step_mode=self.step_mode, quant_params=quant_params)
        elif postpool == 'max':
            self.postpool = sjlayers_t.QuantizedMaxPool(tuple(self._fm_dims), step_mode=self.step_mode, quant_params=quant_params)

        self.flatten = Flatten(step_mode=self.step_mode)
        # Replace Linear with the quantized variant
        self.linear = QuantizedLinear(self.in_planes*QuantizedBasicBlock.expansion,
                                      output_shape[0],
                                      step_mode=self.step_mode,
                                      quant_params=quant_params)

    def _make_layer(self,  # noqa: PLR0913
                    basicblockbuilder: QuantizedBasicBlockBuilder,
                    num_blocks: int,
                    planes: int,
                    kernel_size: int,
                    stride: int,
                    padding: int) -> nn.Sequential:
        strides = [stride] + [1]*(num_blocks-1)
        layers: list[QuantizedBasicBlock] = []
        for stride in strides:
            block: QuantizedBasicBlock = basicblockbuilder(in_planes=self.in_planes,
                                      planes=planes,
                                      kernel_size=kernel_size,
                                      stride=stride,
                                      padding=padding)
            layers.append(block)
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002
        x = input

        out = self.identity1(x)

        if hasattr(self, 'prepool'):
            out = self.prepool(out)

        out = self.conv1(out)
        if self.batch_norm:
            out = self.bn1(out)
        out = self.neuron1(out)

        for i in range(len(self.layers)):
            out = self.layers[i](out)

        if hasattr(self, 'postpool'):
            out = self.postpool(out)

        out = self.flatten(out)
        return cast(torch.Tensor, self.linear(out))



"""Contains quantized spiking neuron implementations."""

from __future__ import annotations

import sys
from typing import Literal

import torch
from qualia_core.learningmodel.pytorch.layers.QuantizedLayer import QuantizedLayer, QuantizerActProtocol, QuantizerInputProtocol
from qualia_core.learningmodel.pytorch.Quantizer import QuantizationConfig, Quantizer, update_params
from spikingjelly.activation_based.neuron import IFNode, LIFNode  # type: ignore[import-untyped]

from .CustomNode import ATIF, SpikeFunctionSigmoid

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override


class QuantizedLIFNode(LIFNode,  # type: ignore[misc]
                       QuantizerInputProtocol,
                       QuantizerActProtocol,
                       QuantizedLayer):
    """Quantized variant of SpikingJelly's :class:`spikingjelly.activation_based.neuron.LIFNode`.

    Hyperparameters ``v_threshold``, ``v_reset`` and ``tau`` are quantized as well as membrane potential ``v``.
    """

    v: torch.Tensor
    v_threshold: float
    v_reset: float | None
    tau: float

    def __init__(self,  # noqa: PLR0913
                 quant_params: QuantizationConfig,
                 tau: float = 2.,
                 decay_input: bool = True,  # noqa: FBT001, FBT002
                 v_threshold: float = 1.,
                 v_reset: float = 0.,
                 detach_reset: bool = False,  # noqa: FBT002, FBT001
                 step_mode: str = 's',
                 backend: str = 'torch') -> None:
        """Construct :class:`QuantizedLIFNode`.

        For more information about spiking neuron parameters, see:
        :meth:`spikingjelly.activation_based.neuron.LIFNode.__init__`

        :param tau: Membrane time constant
        :param decay_input: Whether the input will decay
        :param v_threshold: Threshold of this neurons layer
        :param v_reset: Reset voltage of this neurons layer. If not ``None``, the neuron's voltage will be set to ``v_reset``
                        after firing a spike. If ``None``, the neuron's voltage will subtract ``v_threshold`` after firing a spike
        :param detach_reset: Whether detach the computation graph of reset in backward
        :param step_mode: The step mode, which can be `s` (single-step) or `m` (multi-step)
        :param backend: backend fot this neurons layer, only 'torch' is supported
        :param quant_params: Quantization configuration dict, see :class:`qualia_core.learningmodel.pytorch.Quantizer.Quantizer`
        """
        self.call_super_init = True # Support multiple inheritance from nn.Module
        super().__init__(tau=tau,
                         decay_input=decay_input,
                         v_threshold=v_threshold,
                         v_reset=v_reset,
                         detach_reset=detach_reset,
                         step_mode=step_mode,
                         backend=backend)
        # Create the quantizer instance
        quant_params_input = update_params(tensor_type='input', quant_params=quant_params)
        # Does not work like weights since it's dynamic so need to keep track of global max, don't use 'input' since it can be
        # skipped so use 'act' type
        quant_params_v = update_params(tensor_type='act', quant_params=quant_params)
        quant_params_act = update_params(tensor_type='act', quant_params=quant_params)
        self.quantizer_input = Quantizer(**quant_params_input)
        self.quantizer_v = Quantizer(**quant_params_v)
        self.quantizer_act = Quantizer(**quant_params_act)

    @property
    @override  # type: ignore[misc]
    def supported_backends(self) -> tuple[Literal['torch']]:
        """Supported step_mode and backend.

        Only torch backend is supported.

        :return: Tuple of ``'torch'``
        :raise ValueError: When :attr:`step_mode` is not ``'s'`` or ``'m'``
        """
        if self.step_mode in ['s', 'm']:
            return ('torch',)
        raise ValueError(self.step_mode)

    @override  # type: ignore[misc]
    def neuronal_charge(self, x: torch.Tensor) -> None:
        """Quantized :meth:`spikingjelly.activation_based.neuron.LIFNode.neuronal_charge`.

        Membrane potential and hyperparameters are quantized before and after computation using :meth:`quantize_v_and_hyperparams`.

        :param x: Input tensor
        """
        self.quantize_v_and_hyperparams()
        super().neuronal_charge(x)
        self.quantize_v_and_hyperparams()

    @override  # type: ignore[misc]
    def single_step_forward(self, x: torch.Tensor) -> torch.Tensor:
        """Quantized :meth:`spikingjelly.activation_based.neuron.LIFNode.single_step_forward`.

        Input is (optionally) quantized.
        Membrane potential and hyperparameters are quantized before and after computation using :meth:`quantize_v_and_hyperparams`.

        :param x: Input tensor
        """
        self.v_float_to_tensor(x)
        x = self.quantizer_input(x)

        spike: torch.Tensor

        if self.training:
            self.neuronal_charge(x)
            spike = self.neuronal_fire()
            self.neuronal_reset(spike)
        else:
            self.quantize_v_and_hyperparams()
            if self.v_reset is None:
                if self.decay_input:
                    spike, self.v = self.jit_eval_single_step_forward_soft_reset_decay_input(x,
                                                                                             self.v,
                                                                                             self.v_threshold,
                                                                                             self.tau)
                else:
                    spike, self.v = self.jit_eval_single_step_forward_soft_reset_no_decay_input(x,
                                                                                                self.v,
                                                                                                self.v_threshold,
                                                                                                self.tau)
            elif self.decay_input:
                spike, self.v = self.jit_eval_single_step_forward_hard_reset_decay_input(x,
                                                                                         self.v,
                                                                                         self.v_threshold,
                                                                                         self.v_reset,
                                                                                         self.tau)
            else:
                spike, self.v = self.jit_eval_single_step_forward_hard_reset_no_decay_input(x,
                                                                                            self.v,
                                                                                            self.v_threshold,
                                                                                            self.v_reset,
                                                                                            self.tau)
            self.quantize_v_and_hyperparams()

        return self.quantizer_act(spike)

    @override  # type: ignore[misc]
    def multi_step_forward(self, x_seq: torch.Tensor) -> torch.Tensor:
        """Implement multi-step as loop over single-step for quantized neurons, inefficient but at least it works.

        :param x_seq: Input tensor with timesteps
        :return: Output tensor with timesteps
        """
        out = torch.zeros_like(x_seq)
        for t, d in enumerate(x_seq):
            out[t] = self.single_step_forward(d)
        return out

    @property
    def reciprocal_tau(self) -> float:
        """Return 1 / tau.

        :return: 1 / tau
        """
        return 1 / self.tau

    def quantize_v_and_hyperparams(self) -> None:
        """Quantize potential and hyperparameters (v_threshold, tau, v_reset) in-place with the same quantizer at the same time.

        tau is not quantized directly, instead quantize reciprocal_tau (1 / tau) because this is what is used during inference to
        avoid division.
        """
        self.v, hyperparams = self.quantizer_v(self.v,
                                               bias_tensor=self.get_hyperparams_tensor(device=self.v.device, dtype=self.v.dtype))
        self.v_threshold = hyperparams[0].item()
        self.tau = 1 / hyperparams[1].item() # Turn back reciprocal_tau int tau
        if self.v_reset is not None:
            self.v_reset = hyperparams[2].item()

    def get_hyperparams_tensor(self, device: torch.device, dtype: torch.dtype) -> torch.Tensor:
        """Pack v_threshold, reciprocal_tau and optionally v_reset into the same Tensor.

        :param device: Device to create the tensor on
        :param dtype: Data type for the created tensor
        :return: New tensor with hyperparemeters concatenated
        """
        if self.v_reset is None:
            return torch.tensor([self.v_threshold, self.reciprocal_tau], device=device, dtype=dtype)
        return torch.tensor([self.v_threshold, self.reciprocal_tau, self.v_reset], device=device, dtype=dtype)

    @property
    @override
    def weights_q(self) -> int | None:
        """Number of fractional part bits for the membrane potential and hyperparameters in case of fixed-point quantization.

        See :meth:`qualia_core.learningmodel.pytorch.Quantizer.Quantizer.fractional_bits`.

        :return: Fractional part bits for the membrane potential and hyperparameters or ``None`` if not applicable.
        """
        return self.quantizer_v.fractional_bits

    @property
    @override
    def weights_round_mode(self) -> str | None:
        return self.quantizer_v.roundtype

class QuantizedIFNode(IFNode,  # type: ignore[misc]
                      QuantizerInputProtocol,
                      QuantizerActProtocol,
                      QuantizedLayer):
    """Quantized variant of SpikingJelly's :class:`spikingjelly.activation_based.neuron.IFNode`.

    Hyperparameters ``v_threshold`` and ``v_reset`` are quantized as well as membrane potential ``v``.
    """

    v: torch.Tensor
    v_threshold: float
    v_reset: float | None

    def __init__(self,  # noqa: PLR0913
                 quant_params: QuantizationConfig,
                 v_threshold: float = 1.,
                 v_reset: float = 0.,
                 detach_reset: bool = False,  # noqa: FBT001, FBT002
                 step_mode: str = 's',
                 backend: str = 'torch') -> None:
        """Construct :class:`QuantizedIFNode`.

        For more information about spiking neuron parameters, see:
        :meth:`spikingjelly.activation_based.neuron.IFNode.__init__`

        :param v_threshold: Threshold of this neurons layer
        :param v_reset: Reset voltage of this neurons layer. If not ``None``, the neuron's voltage will be set to ``v_reset``
                        after firing a spike. If ``None``, the neuron's voltage will subtract ``v_threshold`` after firing a spike
        :param detach_reset: Whether detach the computation graph of reset in backward
        :param step_mode: The step mode, which can be `s` (single-step) or `m` (multi-step)
        :param backend: backend fot this neurons layer, only 'torch' is supported
        :param quant_params: Quantization configuration dict, see :class:`qualia_core.learningmodel.pytorch.Quantizer.Quantizer`
        """
        self.call_super_init = True # Support multiple inheritance from nn.Module
        super().__init__(v_threshold=v_threshold,
                         v_reset=v_reset,
                         detach_reset=detach_reset,
                         step_mode=step_mode,
                         backend=backend)
        # Create the quantizer instance
        quant_params_input = update_params(tensor_type = 'input', quant_params = quant_params)
        # Does not work like weights since it's dynamic so need to keep track of global max, don't use 'input' since it can be
        # skipped so use 'act' type
        quant_params_v = update_params(tensor_type='act', quant_params=quant_params)
        quant_params_act = update_params(tensor_type='act', quant_params=quant_params)
        self.quantizer_input = Quantizer(**quant_params_input)
        self.quantizer_v = Quantizer(**quant_params_v)
        self.quantizer_act = Quantizer(**quant_params_act)

    @property
    @override  # type: ignore[misc]
    def supported_backends(self) -> tuple[Literal['torch']]:
        """Supported step_mode and backend.

        Only torch backend is supported.

        :return: Tuple of ``'torch'``
        :raise ValueError: When :attr:`step_mode` is not ``'s'`` or ``'m'``
        """
        if self.step_mode in ['s', 'm']:
            return ('torch',)
        raise ValueError(self.step_mode)

    @override  # type: ignore[misc]
    def neuronal_charge(self, x: torch.Tensor) -> None:
        """Quantized :meth:`spikingjelly.activation_based.neuron.IFNode.neuronal_charge`.

        Membrane potential and hyperparameters are quantized before and after computation using :meth:`quantize_v_and_hyperparams`.

        :param x: Input tensor
        """
        self.quantize_v_and_hyperparams()
        super().neuronal_charge(x)
        self.quantize_v_and_hyperparams()

    @override  # type: ignore[misc]
    def single_step_forward(self, x: torch.Tensor) -> torch.Tensor:
        """Quantized :meth:`spikingjelly.activation_based.neuron.IFNode.single_step_forward`.

        Input is (optionally) quantized.
        Membrane potential and hyperparameters are quantized before and after computation using :meth:`quantize_v_and_hyperparams`.

        :param x: Input tensor
        """
        self.v_float_to_tensor(x)
        x = self.quantizer_input(x)

        spike: torch.Tensor

        if self.training:
            self.neuronal_charge(x)
            spike = self.neuronal_fire()
            self.neuronal_reset(spike)
        else:
            self.quantize_v_and_hyperparams()
            if self.v_reset is None:
                spike, self.v = self.jit_eval_single_step_forward_soft_reset(x, self.v, self.v_threshold)
            else:
                spike, self.v = self.jit_eval_single_step_forward_hard_reset(x, self.v, self.v_threshold, self.v_reset)
            self.quantize_v_and_hyperparams()

        return self.quantizer_act(spike)

    @override  # type: ignore[misc]
    def multi_step_forward(self, x_seq: torch.Tensor) -> torch.Tensor:
        """Implement multi-step as loop over single-step for quantized neurons, inefficient but at least it works.

        :param x_seq: Input tensor with timesteps
        :return: Output tensor with timesteps
        """
        out = torch.zeros_like(x_seq)
        for t, d in enumerate(x_seq):
            out[t] = self.single_step_forward(d)
        return out

    def quantize_v_and_hyperparams(self) -> None:
        """Quantize potential and hyperparameters (v_threshold, v_reset) in-place with the same quantizer at the same time."""
        self.v, hyperparams = self.quantizer_v(self.v,
                                               bias_tensor=self.get_hyperparams_tensor(device=self.v.device, dtype=self.v.dtype))
        self.v_threshold = hyperparams[0].item()
        if self.v_reset is not None:
            self.v_reset = hyperparams[1].item()

    def get_hyperparams_tensor(self, device: torch.device, dtype: torch.dtype) -> torch.Tensor:
        """Pack v_threshold and optionally v_reset into the same Tensor.

        :param device: Device to create the tensor on
        :param dtype: Data type for the created tensor
        :return: New tensor with hyperparemeters concatenated
        """
        if self.v_reset is None:
            return torch.tensor([self.v_threshold], device=device, dtype=dtype)
        return torch.tensor([self.v_threshold, self.v_reset], device=device, dtype=dtype)

    @property
    @override
    def weights_q(self) -> int | None:
        """Number of fractional part bits for the membrane potential and hyperparameters in case of fixed-point quantization.

        See :meth:`qualia_core.learningmodel.pytorch.Quantizer.Quantizer.fractional_bits`.

        :return: Fractional part bits for the membrane potential and hyperparameters or ``None`` if not applicable.
        """
        return self.quantizer_v.fractional_bits

    @property
    @override
    def weights_round_mode(self) -> str | None:
        return self.quantizer_v.roundtype

class QuantizedATIF(ATIF,
                    QuantizerInputProtocol,
                    QuantizerActProtocol,
                    QuantizedLayer):
    """Quantized Integrate and Fire soft-reset with learnable Vth and activation scaling, based on spikingjelly."""

    def __init__(self,  # noqa: PLR0913
                 quant_params: QuantizationConfig,
                 v_threshold: float = 1.0,
                 vth_init_l: float = 0.8,
                 vth_init_h: float = 1.,
                 alpha: float = 1.,
                 device: str = 'cpu') -> None:
        """Construct :class:`ATIF`.

        :param v_threshold: Factor to apply to the uniform initialization bounds
        :param vth_init_l: Lower bound for uniform initialization of threshold Tensor
        :param vth_init_h: Higher bound for uniform initialization of threshold Tensor
        :param alpha: Sigmoig surrogate scale factor
        :param device: Device to run the computation on
        :param quant_params: Quantization configuration dict, see :class:`qualia_core.learningmodel.pytorch.Quantizer.Quantizer`
        """
        super().__init__(v_threshold=v_threshold,
                         vth_init_l=vth_init_l,
                         vth_init_h=vth_init_h,
                         alpha=alpha,
                         device=device)
        # Create the quantizer instance
        quant_params_input = update_params(tensor_type='input', quant_params=quant_params)
        # Does not work like weights since it's dynamic so need to keep track of global max, don't use 'input' since it can be
        # skipped so use 'act' type
        quant_params_v = update_params(tensor_type='act', quant_params=quant_params)
        quant_params_act = update_params(tensor_type='act', quant_params=quant_params)
        self.quantizer_input = Quantizer(**quant_params_input)
        self.quantizer_v = Quantizer(**quant_params_v)
        self.quantizer_act = Quantizer(**quant_params_act)

    @override
    def ifsrl_fn(self, x: torch.Tensor) -> torch.Tensor:
        """Quantized :meth:`qualia_plugin_snn.learningmodel.pytorch.layers.CustomNode.ATIF.ifsrl_fn`.

        Input is quantized.
        Membrane potential is quantized before and after computation.
        Threshold is quantized by :meth:`get_coeffs`.

        :param x: Input tensor
        :return: Output tensor
        """
        # Primary membrane charge
        self.v_float_to_tensor(x)
        x = self.quantizer_input(x)
        self.v = self.quantizer_v(self.v)
        self.v = self.v + x
        self.v = self.quantizer_v(self.v)

        # Fire
        q = self.quantizer_v(self.v - self.vp_th)

        z = SpikeFunctionSigmoid.apply(q, self.alpha * torch.ones(1).to(self.device)).float()

        # Soft-Reset
        self.v = (1. - z) * self.v + z * (self.v - self.vp_th)
        self.v = self.quantizer_v(self.v)
        return self.quantizer_act(z * self.get_coeffs())

    @override
    def get_coeffs(self) -> torch.Tensor:
        """Return the quantized Tensor of threshold :attr:`v_threshold`.

        :return: Quantized Tensor of threshold :attr:`v_threshold`
        """
        return self.quantizer_v(self.v_threshold)

    @override
    def set_coeffs(self, v_threshold: torch.Tensor) -> None:
        """Quantized and replace the Tensor of threshold :attr:`v_threshold`.

        :param v_threshold: New Tensor of quantized threshold to replace :attr:`v_threshold`
        """
        _ = self.v_threshold.copy_(self.quantizer_v(v_threshold))

    @property
    @override
    def weights_q(self) -> int | None:
        """Number of fractional part bits for the membrane potential and hyperparameters in case of fixed-point quantization.

        See :meth:`qualia_core.learningmodel.pytorch.Quantizer.Quantizer.fractional_bits`.

        :return: Fractional part bits for the membrane potential and hyperparameters or ``None`` if not applicable.
        """
        return self.quantizer_v.fractional_bits

    @property
    @override
    def weights_round_mode(self) -> str | None:
        return self.quantizer_v.roundtype



"""Contains the base class for spiking neural network models."""

from __future__ import annotations

import logging

import spikingjelly.activation_based.neuron as sj  # type: ignore[import-untyped]
from qualia_core.learningmodel.pytorch.LearningModelPyTorch import LearningModelPyTorch
from qualia_core.typing import TYPE_CHECKING

import qualia_plugin_snn.learningmodel.pytorch.layers.quantized_SNN_layers as qsj
from qualia_plugin_snn.learningmodel.pytorch.layers import CustomNode

if TYPE_CHECKING:
    from qualia_core.learningmodel.pytorch.Quantizer import QuantizationConfig  # noqa: TCH002
    from qualia_core.typing import RecursiveConfigDict, RecursiveConfigUnion
    from torch import nn  # noqa: TCH002

logger = logging.getLogger(__name__)

class SNN(LearningModelPyTorch):
    """Base class for spiking neural network models to inherit from."""

    timesteps: int #: Number of timesteps
    step_mode: str #: SpikingJelly ``step_mode``
    is_snn: bool = True #: Always ``True`` in case of spiking neural networks

    def __select_neuron(self, neuron: RecursiveConfigDict) -> type[nn.Module]:
        """Select a spiking neuron class from a kind specified in configuration file.

        The class will be looked up by its name in the following modules successively, first match is used:

            #. :mod:`spikingjelly.activation_based.neuron`
            #. :mod:`qualia_plugin_snn.learningmodel.pytorch.layers.CustomNode`
            #. :mod:`qualia_plugin_snn.learningmodel.pytorch.quantized_SNN_layers`

        :meta public:
        :param neuron: A spiking neuron configuration dict
        :return: The class corresponding to the ``kind`` specified in the neuron configuration dict
        :raise ValueError: When the ``'kind'`` key cannot be found in the ``neuron`` dict
        :raise TypeError: When the value associated to ``kind`` is not a string
        :raise AttributeError: When the class for ``kind`` cannot be looked up
        """
        # SpikeNeuron Selection
        if 'kind' not in neuron:
            logger.error('`params.neuron.kind` is required')
            raise ValueError
        if not isinstance(neuron['kind'], str):
            logger.error('`params.neuron.kind` must be a string, got: %s', type(neuron['kind']))
            raise TypeError
        neurons_kind = neuron['kind']
        neuron_type: type[nn.Module]

        try:
            neuron_type = getattr(sj, neurons_kind)
        except AttributeError:
            logger.info("Module 'spikingjelly.activation_based.neuron' has no attribute %s",  neurons_kind)
            logger.info('Checking learningmodel/pytorch/layers/CustomNode')
            try:
                neuron_type = getattr(CustomNode, neurons_kind)
            except AttributeError:
                logger.info("Module 'learningmodel.pytorch.layers.CustomNode' has no attribute %s", neurons_kind)
                logger.info('Checking learningmodel/pytorch/Quantized_SNN_layers')
                try:
                    neuron_type = getattr(qsj, neurons_kind)
                except AttributeError:
                    logger.exception("Module 'learningmodel.pytorch.Quantized_SNN_layers' has no attribute %s", neurons_kind)
                    raise

        return neuron_type

    def __extract_neuron_params(self, neuron: RecursiveConfigDict) -> dict[str, RecursiveConfigUnion | None]:
        """Extract params from the given ``neuron`` configuration dict, also convert ``v_reset``.

        ``v_reset`` is specified as either a float (hard-reset) or ``false`` (soft-reset) in the TOML configuration since there is
        no ``None`` equivalent in TOML. However SpikingJelly neurons use ``None`` to signify soft-reset, so convert
        ``v_reset=false`` to ``v_reset=None``.

        :meta public:
        :param neuron: The spiking neuron configuration dict
        :return: The ``params`` dict for the given neuron configuration dict with converted ``v_reset``
        :raise TypeError: When the value associated to ``params`` is not a dict
        """
        # Extract neuron params
        neuron_params = neuron.get('params', {})
        if not isinstance(neuron_params, dict):
            logger.error('`params.neuron.params` must be a dict, got: %s', type(neuron_params))
            raise TypeError

        # Replace v_reset = False with v_reset = None for soft reset of SpikingJelly
        filtered_neuron_params: dict[str, RecursiveConfigUnion | None] = {
                k: v for k, v in neuron_params.items()
                if k != 'v_reset'
                }

        if 'v_reset' in neuron_params:
            if neuron_params['v_reset'] is False:
                filtered_neuron_params['v_reset'] = None
            else:
                filtered_neuron_params['v_reset'] = neuron_params['v_reset']

        return filtered_neuron_params

    def create_neuron(self, quant_params: QuantizationConfig | None = None) -> nn.Module:
        """Instanciate a spiking neuron from the kind and params found in ``neuron`` of :meth:``__init__``.

        :param quant_params: Optional quantization configuration dict in case of quantized network, see
                             :class:`qualia_core.learningmodel.pytorch.Quantizer.Quantizer`
        :return: A spiking neuron instance
        """
        if quant_params is not None:
            return self.__neuron_type(**self.__neuron_params, quant_params=quant_params)
        return self.__neuron_type(**self.__neuron_params)

    def __init__(self,
                 input_shape: tuple[int, ...],
                 output_shape: tuple[int, ...],
                 timesteps: int,
                 neuron: RecursiveConfigDict | None) -> None:
        """Construct :class:`SNN`.

        :attr:`step_mode` is extracted from the neuron configuration dict ``neuron``.
        :meth:`__select_neuron` and :meth:`__extract_neuron_params` are called to prepare the neuron for instanciation with
        :meth:`create_neuron` which should be used in derived classes to instanciate the spiking neurons.

        Below is an example of a neuron configuration dict for SpikingJelly's multi-step
        :class:`spikingjelly.activation_based.neuron.IFNode` soft-reset with threshold at 1.0.

        * Python

        .. code-block:: python

            neuron = {
                'kind': 'IFNode',
                'params': {
                    'v_threshold': 1.0,
                    'v_reset': False,
                    'step_mode': 'm',
                    }
            }

        * TOML

        .. code-block:: toml

            [[model]]
            params.neuron.kind = 'IFNode'
            params.neuron.params.v_threshold = 1.0
            params.neuron.params.v_reset = false
            params.neuron.params.step_mode = 'm'

        :param input_shape: Input shape passed to :class:`qualia_core.learningmodel.pytorch.LearningModel.LearningModel`
                            with ``timesteps`` prepended
        :param output_shape: Output shape passed to :class:`qualia_core.learningmodel.pytorch.LearningModel.LearningModel`
        :param timesteps: Number of timesteps
        :param neuron: Spiking neuron configuration dict
        :raise ValueError: When ``neuron`` is ``None`` or empty
        :raise TypeError: When the value associated to ``params.step_mode`` is not a string
        """
        super().__init__(input_shape=(timesteps, *tuple(input_shape)), output_shape=output_shape)
        self.timesteps = timesteps

        if not neuron:
            logger.error('`params.neuron` is required')
            raise ValueError

        # Find neuron class
        self.__neuron_type = self.__select_neuron(neuron)

        self.__neuron_params = self.__extract_neuron_params(neuron)

        # Extract step_mode
        step_mode = self.__neuron_params.get('step_mode', 's')
        if not isinstance(step_mode, str):
            logger.error('`params.neuron.params.step_mode` must be a string, got: %s', type(step_mode))
            raise TypeError
        self.step_mode = step_mode




"""Provide the EnergyEstimationMetric postprocessing module based on Lemaire et al., 2022."""

from __future__ import annotations

import dataclasses
import functools
import logging
import math
import sys
from collections.abc import Generator
from dataclasses import dataclass
from typing import Any, Callable, NamedTuple, cast

import torch
from qualia_core.learningmodel.pytorch.layers import Add
from qualia_core.learningmodel.pytorch.layers.quantized_layers import QuantizedIdentity
from qualia_core.learningmodel.pytorch.layers.QuantizedAdd import QuantizedAdd
from qualia_core.learningmodel.pytorch.Quantizer import Quantizer
from qualia_core.postprocessing.PostProcessing import PostProcessing
from qualia_core.typing import TYPE_CHECKING, ModelConfigDict
from qualia_core.utils.logger import CSVLogger
from torch import nn

from qualia_plugin_snn.learningframework.SpikingJelly import SpikingJelly
from qualia_plugin_snn.learningmodel.pytorch.SNN import SNN

# We are inside a TYPE_CHECKING block but our custom TYPE_CHECKING constant triggers TCH001-TCH003 so ignore them
if TYPE_CHECKING:
    from qualia_codegen_core.graph import ModelGraph  # noqa: TCH002
    from qualia_codegen_core.graph.layers import TBaseLayer, TConvLayer, TDenseLayer  # noqa: TCH002
    from qualia_core.datamodel.RawDataModel import RawData  # noqa: TCH002
    from qualia_core.qualia import TrainResult  # noqa: TCH002
    from torch.types import Number  # noqa: TCH002

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class EnergyEstimationMetricLoggerFields(NamedTuple):
    """Interface object for CSV logging.

    Should contain the same fields as :class:`EnergyMetrics` and returned by :meth:`EnergyMetrics.asnamedtuple`.

    :param name: Layer name
    :param mem_pot: Energy for potential memory read/write
    :param mem_weights: Energy for weights memory read
    :param mem_bias: Energy for bias memory read
    :param mem_io: Energy for input/output memory read/write
    :param ops: Energy for synaptic operations
    :param addr: Energy for event address computation
    :param input_spikerate: Average input spike rate per timestep
    :param output_spikerate: Average output spike rate per timestep
    :param input_is_binary: If input tensor only contains binary values, i.e., spikes
    :param output_is_binary: If output tensor only contains binary values, i.e., spikes
    """

    #: Layer name
    name: str
    #: Energy for potential memory read/write
    mem_pot: float
    #: Energy for weights memory read
    mem_weights:float
    #: Energy for bias memory read
    mem_bias: float
    #: Energy for input/output memory read/write
    mem_io: float
    #: Energy for potential, weights, bias and input/output read/write, i.e., Erdram + Ewrram
    mem_total: float
    #: Energy for synaptic operations
    ops: float
    #: Energy for event address computation
    addr: float
    #: Energy for synaptic operations and event address computation, i.e., Eops+addr
    opsaddr: float
    #: Total energy, i.e., energy for memory read/write, synaptic operations and event adress computation
    total: float
    #: Average input spike rate per timestep
    input_spikerate: float
    #: Average output spike rate per timestep
    output_spikerate: float
    #: If input tensor only contains binary values, i.e., spikes
    input_is_binary: bool
    #: If output tensor only contains binary values, i.e., spikes
    output_is_binary: bool

@dataclass
class EnergyMetrics:
    """Holds the computed average energy per inference for each layer.

    :param name: Layer name
    :param mem_pot: Energy for potential memory read/write
    :param mem_weights: Energy for weights memory read
    :param mem_bias: Energy for bias memory read
    :param mem_io: Energy for input/output memory read/write
    :param ops: Energy for synaptic operations
    :param addr: Energy for event address computation
    :param input_spikerate: Average input spike rate per timestep
    :param output_spikerate: Average output spike rate per timestep
    :param input_is_binary: If input tensor only contains binary values, i.e., spikes
    :param output_is_binary: If output tensor only contains binary values, i.e., spikes
    """

    #: Layer name
    name: str
    #: Energy for potential memory read/write
    mem_pot: float
    #: Energy for weights memory read
    mem_weights: float
    #: Energy for bias memory read
    mem_bias: float
    #: Energy for input/output memory read/write
    mem_io: float
    #: Energy for synaptic operations
    ops: float
    #: Energy for event address computation
    addr: float
    #: Average input spike rate per timestep
    input_spikerate: float | None
    #: Average output spike rate per timestep
    output_spikerate: float | None
    #: If input tensor only contains binary values, i.e., spikes
    input_is_binary: bool
    #: If output tensor only contains binary values, i.e., spikes
    output_is_binary: bool

    @property
    def mem_total(self) -> float:
        """Energy for potential, weights, bias and input/output read/write. Erdram + Ewrram."""
        return self.mem_pot + self.mem_weights + self.mem_bias + self.mem_io

    @property
    def opsaddr(self) -> float:
        """Energy for synaptic operations and event address computation. Eops+addr."""
        return self.ops + self.addr

    @property
    def total(self) -> float:
        """Total energy. Energy for memory read/write, synaptic operations and event adress computation."""
        return self.mem_total + self.opsaddr

    def asnamedtuple(self) -> EnergyEstimationMetricLoggerFields:
        """Return the data from this class as a NamedTuple for use with the CSV logger.

        Instanciate a :class:`EnergyEstimationMetricLoggerFields` object with
        all of this class fields and properties and return it.

        :return: the :class:`EnergyEstimationMetricLoggerFields` with all data from this object copied into it
        """
        return EnergyEstimationMetricLoggerFields(**dataclasses.asdict(self),
                                                  mem_total=self.mem_total,
                                                  opsaddr=self.opsaddr,
                                                  total=self.total)

class EnergyEstimationMetric(PostProcessing[nn.Module]):
    r"""Analytical energy estimation metric.

    From `An Analytical Estimation of Spiking Neural Networks Energy Efficiency <https://arxiv.org/abs/2210.13107>`_,
    Lemaire et al. ICONIP2022.

    .. code-block:: bibtex

        @inproceedings{EnergyEstimationMetricICONIP2022,
            title = {An Analytical Estimation of Spiking Neural Networks Energy Efficiency},
            author = {Lemaire, Edgar and Cordone, Loïc and Castagnetti, Andrea
                      and Novac, Pierre-Emmanuel and Courtois, Jonathan and Miramond, Benoît},
            booktitle = {Proceedings of the 29th International Conference on Neural Information Processing},
            pages = {574--587},
            year = {2023},
            doi = {10.1007/978-3-031-30105-6_48},
            series = {ICONIP},
        }

    Supports sequential (non-residual) formal and spiking convolutional neural networks with the following layers:

    * :class:`torch.nn.Conv1d`
    * :class:`torch.nn.Conv2d`
    * :class:`torch.nn.Linear`
    * :class:`torch.nn.ReLU` for formal neural networks
    * :class:`spikingjelly.activation_based.neuron.IFNode` for spiking neural networks
    * :class:`spikingjelly.activation_based.neuron.LIFNode` for spiking neural networks
    """

    _m_e_add: float = 0.1
    """Energy for add operation in pJ, 45nm int32.

    Default from `Computing’s Energy Problem (and what we can do about it) <https://gwern.net/doc/cs/hardware/2014-horowitz-2.pdf>`_,
    Mark Horowitz, ISSCC 2014.

    :meta public:
    """  # noqa: RUF001

    _m_e_mul: float = 3.1
    """Energy for mul operation in pJ, 45nm int32.

    Default from `Computing’s Energy Problem (and what we can do about it) <https://gwern.net/doc/cs/hardware/2014-horowitz-2.pdf>`_,
    Mark Horowitz, ISSCC 2014.

    :meta public:
    """  # noqa: RUF001

    #: CSV logger to record metrics for each layer in a file inside the `logs/<bench.name>/EnergyEstimationMetric` directory
    csvlogger: CSVLogger[EnergyEstimationMetricLoggerFields]

    def __init__(self,
                 mem_width: int,
                 fifo_size: int = 0,
                 total_spikerate_exclude_nonbinary: bool = True) -> None:  # noqa: FBT001, FBT002
        """Construct :class:`qualia_plugin_snn.postprocessing.EnergyEstimationMetric.EnergyEstimationMetric`.

        :param mem_width: Memory access size in bits, e.g. 16 for 16-bit quantization
        :param fifo_size: Size of the input/output FIFOs for each layer in SPLEAT
        :param total_spikerate_exclude_nonbinary: If True, exclude non-binary inputs/outputs from total spikerate computation
        """
        super().__init__()
        self._mem_width = mem_width
        self._fifo_size = fifo_size
        self._total_spikerate_exclude_nonbinary = total_spikerate_exclude_nonbinary

        self.csvlogger = CSVLogger(name='EnergyEstimationMetric')
        self.csvlogger.fields = EnergyEstimationMetricLoggerFields

    #######
    # FNN #
    #######

    def _rdin_conv_fnn(self, layer: TConvLayer) -> int:
        """Count number of read operations for the input of a convolutional layer in a formal neural network.

        Eq. 3: Cin × Cout × Hout × Wout × Wkernel × Hkernel.

        :meta public:
        :param layer: A convolutional layer
        :return: Number of read operations for the input of a convolutional layer in a formal neural network
        """  # noqa: RUF002
        return layer.input_shape[0][-1] * math.prod(layer.output_shape[0][1:]) * math.prod(layer.kernel_size)

    def _e_rdin_conv_fnn(self, layer: TConvLayer, e_rdram: Callable[[int], float]) -> float:
        """Compute energy for read operations for the input of a convolution layer in a formal neural network.

        :meta public:
        :param layer: A convolutional layer
        :param e_rdram: Function to compute memory read access energy for a given memory size
        :return: Energy for read operations for the input of a convolution layer in a formal neural network
        """
        return self._rdin_conv_fnn(layer) * e_rdram(math.prod(layer.input_shape[0][1:]))


    def _rdin_fc_fnn(self, layer: TDenseLayer) -> int:
        """Count number of read operations for the input of a fully-connected layer in a formal neural network.

        Eq. 4: Nin.

        :meta public:
        :param layer: A fully-connected layer
        :return: Number of read operations for the input of a fully-connected layer in a formal neural network
        """
        return layer.input_shape[0][-1]

    def _e_rdin_fc_fnn(self, layer: TDenseLayer, e_rdram: Callable[[int], float]) -> float:
        """Compute energy for read operations for the input of a fully-connected layer in a formal neural network.

        :meta public:
        :param layer: A fully-connected layer
        :param e_rdram: Function to compute memory read access energy for a given memory size
        :return: Energy for read operations for the input of a fully-connected layer in a formal neural network
        """
        return self._rdin_fc_fnn(layer) * e_rdram(math.prod(layer.input_shape[0][1:]))


    def _rdweights_conv_fnn(self, layer: TConvLayer) -> int:
        """Count number of read operations for the weights of a convolutional layer (excluding bias) in a formal neural network.

        Eq. 6.1f: Cin × Wkernel × Hkernel × Cout × Wout × Hout.

        :meta public:
        :param layer: A convolutional layer
        :return: Number of read operations for the weights of a convolutional layer (excluding bias) in a formal neural network
        """  # noqa: RUF002
        return layer.input_shape[0][-1] * math.prod(layer.kernel_size) * math.prod(layer.output_shape[0][1:])

    def _e_rdweights_conv_fnn(self, layer: TConvLayer, e_rdram: Callable[[int], float]) -> float:
        """Compute energy for read operations for the weights of a convolutional layer (excluding bias) in a formal neural network.

        :meta public:
        :param layer: A convolutional layer
        :param e_rdram: Function to compute memory read access energy for a given memory size
        :return: Energy for read operations for the weights of a convolutional layer (excluding bias) in a formal neural network
        """
        return self._rdweights_conv_fnn(layer) * e_rdram(layer.kernel.size)

    def _rdbias_conv_fnn(self, layer: TConvLayer) -> int:
        """Count number of read operations for the biases of a convolutional layer in a formal neural network.

        Eq. 6.1s: Cout × Wout × Hout.

        :meta public:
        :param layer: A convolutional layer
        :return: Number of read operations for the biases of a convolutional layer in a formal neural network
        """  # noqa: RUF002
        return math.prod(layer.output_shape[0][1:])

    def _e_rdbias_conv_fnn(self, layer: TConvLayer, e_rdram: Callable[[int], float]) -> float:
        """Compute energy for read operations for the biases of a convolutional layer in a formal neural network.

        :meta public:
        :param layer: A convolutional layer
        :param e_rdram: Function to compute memory read access energy for a given memory size
        :return: Energy for read operations for the biases of a convolutional layer in a formal neural network or 0 if the layer
                 does not use biases
        """
        if layer.use_bias:
            return self._rdbias_conv_fnn(layer) * e_rdram(layer.bias.size)
        return 0


    def _rdweights_fc_fnn(self, layer: TDenseLayer) -> int:
        """Count number of read operations for the weights of a fully-connected layer (excluding bias) in a formal neural network.

        Eq. 6.2f: Nin × Nout.

        :meta public:
        :param layer: A fully-connected layer
        :return: Number of read operations for the weights of a fully-connected layer (excluding bias) in a formal neural network
        """  # noqa: RUF002
        return layer.input_shape[0][-1] * layer.output_shape[0][-1]

    def _e_rdweights_fc_fnn(self, layer: TDenseLayer, e_rdram: Callable[[int], float]) -> float:
        """Compute energy for read operations for weights of a fully-connected layer (excluding bias) in a formal neural network.

        :meta public:
        :param layer: A fully-connected layer
        :param e_rdram: Function to compute memory read access energy for a given memory size
        :return: Energy for read operations for the weights of a fully-connected layer (excluding bias) in a formal neural network
        """
        return self._rdweights_fc_fnn(layer) * e_rdram(layer.kernel.size)

    def _rdbias_fc_fnn(self, layer: TDenseLayer) -> int:
        """Count number of read operations for the biases of a fully-connected layer in a formal neural network.

        Eq. 6.2s: Nout.

        :meta public:
        :param layer: A fully-connected layer
        :return: Number of read operations for the biases of a fully-connected layer in a formal neural network
        """
        return layer.output_shape[0][-1]

    def _e_rdbias_fc_fnn(self, layer: TDenseLayer, e_rdram: Callable[[int], float]) -> float:
        """Compute energy for read operations for the biases of a fully-connected layer in a formal neural network.

        :meta public:
        :param layer: A fully-connected layer
        :param e_rdram: Function to compute memory read access energy for a given memory size
        :return: Energy for read operations for the biases of a fully-connected layer in a formal neural network or 0 if the layer
                 does not use biases
        """
        if layer.use_bias:
            return self._rdbias_fc_fnn(layer) * e_rdram(layer.bias.size)
        return 0


    def _wrout_conv_fnn(self, layer: TConvLayer) -> int:
        """Count number of write operations for the output of a convolutional layer in a formal neural network.

        Eq. 11: Cout × Hout × Wout.

        :meta public:
        :param layer: A convolutional layer
        :return: Number of write operations for the output of a convolutional layer in a formal neural network
        """  # noqa: RUF002
        return math.prod(layer.output_shape[0][1:])

    def _e_wrout_conv_fnn(self, layer: TConvLayer, e_wrram: Callable[[int], float]) -> float:
        """Compute energy for write operations for the output of a convolutional layer in a formal neural network.

        :meta public:
        :param layer: A convolutional layer
        :param e_rdram: Function to compute memory write access energy for a given memory size
        :return: Energy for write operations for the output of a convolutional layer in a formal neural network
        """
        return self._wrout_conv_fnn(layer) * e_wrram(math.prod(layer.output_shape[0][1:]))


    def _wrout_fc_fnn(self, layer: TDenseLayer) -> int:
        """Count number of write operations for the output of a fully-connected layer in a formal neural network.

        Eq. 12: Nout.

        :meta public:
        :param layer: A fully-connected layer
        :return: Number of write operations for the output of a fully-connected layer in a formal neural network
        """
        return layer.output_shape[0][-1]

    def _e_wrout_fc_fnn(self, layer: TDenseLayer, e_wrram: Callable[[int], float]) -> float:
        """Compute energy for write operations for the output of a fully-connected layer in a formal neural network.

        :meta public:
        :param layer: A fully-connected layer
        :param e_rdram: Function to compute memory write access energy for a given memory size
        :return: Energy for write operations for the output of a fully-connected layer in a formal neural network
        """
        return self._wrout_fc_fnn(layer) * e_wrram(math.prod(layer.output_shape[0][1:]))


    def _mac_ops_conv_fnn(self, layer: TConvLayer) -> int:
        """Count number of multiply-accumulate operations inside a convolutional layer in a formal neural network.

        Eq. 1.1: Cout × Hout × Wout × Cin × Hkernel × Wkernel.

        :meta public:
        :param layer: A convolutional layer
        :return: Number of multiply-accumulate operations inside a convolutional layer in a formal neural network.
        """  # noqa: RUF002
        return math.prod(layer.output_shape[0][1:]) * layer.input_shape[0][-1] * math.prod(layer.kernel_size)

    def _acc_ops_conv_fnn(self, layer: TConvLayer) -> int:
        """Count number of accumulate operations inside a convolutional layer in a formal neural network.

        Eq. 1.2: Cout × Hout × Wout.

        :meta public
        :param layer: A convolutional layer
        :return: Number of accumulate operations inside a convolutional layer in a formal neural network.
        """  # noqa: RUF002
        return math.prod(layer.output_shape[0][1:])

    def _e_ops_conv_fnn(self, layer: TConvLayer) -> float:
        """Compute energy for multiply-accumulate and accumulate operations inside convolutional layer in a formal neural network.

        :meta public:
        :param layer: A convolutional layer
        :return: Energy for multiply-accumulate and accumulate operations inside a convolutional layer in a formal neural network.
        """
        return self._mac_ops_conv_fnn(layer) * (self._e_mul + self._e_add) + self._acc_ops_conv_fnn(layer) * self._e_add


    def _mac_ops_fc_fnn(self, layer: TDenseLayer) -> int:
        """Count number of multiply-accumulate operations inside a fully-connected layer in a formal neural network.

        Eq. 2.1: Nin × Nout.

        :meta public:
        :param layer: A fully-connected layer
        :return: Number of multiply-accumulate operations inside a fully-connected layer in a formal neural network.
        """  # noqa: RUF002
        return layer.input_shape[0][-1] * layer.output_shape[0][-1]

    def _acc_ops_fc_fnn(self, layer: TDenseLayer) -> int:
        """Count number of accumulate operations inside a fully-connected layer in a formal neural network.

        Eq. 2.2: Nout.

        :meta public:
        :param layer: A fully-connected layer
        :return: Number of accumulate operations inside a fully-connected layer in a formal neural network.
        """
        return layer.output_shape[0][-1]

    def _e_ops_fc_fnn(self, layer: TDenseLayer) -> float:
        """Compute energy for multiply-accumulate and accumulate operations inside fully-connected layer in formal neural network.

        :meta public:
        :param layer: A fully-connected layer
        :return: Energy for multiply-accumulate and accumulate operations inside a fully-connected layer in formal neural network.
        """
        return self._mac_ops_fc_fnn(layer) * (self._e_mul + self._e_add) + self._acc_ops_fc_fnn(layer) * self._e_add


    def _mac_addr_conv_fnn(self, _: TConvLayer) -> int:
        """Count number of multiply-accumulate operations for addressing a convolutional layer in a formal neural network.

        0

        :meta public:
        :return: 0
        """
        return 0

    def _acc_addr_conv_fnn(self, layer: TConvLayer) -> int:
        """Count number of accumulate operations for addressing a convolutional layer in a formal neural network.

        Cin × Hin × Win + Cout × Hout × Wout + Cout × Hkernel × Wkernel.

        :meta public:
        :param layer: A convolutional layer
        :return: Number of accumulate operations for addressing a convolutional layer in a formal neural network
        """  # noqa: RUF002
        return (math.prod(layer.input_shape[0][1:])
                + math.prod(layer.output_shape[0][1:])
                + layer.output_shape[0][-1] * math.prod(layer.kernel_size))

    def _e_addr_conv_fnn(self, layer: TConvLayer) -> float:
        """Compute energy for addressing a convolutional layer in a formal neural network.

        :meta public:
        :param layer: A convolutional layer
        :return: Energy for addressing a convolutional layer in a formal neural network.
        """
        return self._mac_addr_conv_fnn(layer) * (self._e_mul + self._e_add) + self._acc_addr_conv_fnn(layer) * self._e_add


    def _mac_addr_fc_fnn(self, _: TDenseLayer) -> int:
        """Count number of multiply-accumulate operations for addressing a fully-connected layer in a formal neural network.

        0

        :meta public:
        :return: 0
        """
        return 0

    def _acc_addr_fc_fnn(self, layer: TDenseLayer) -> int:
        """Count number of accumulate operations for addressing a fully-connected layer in a formal neural network.

        Nin × Nout.

        :meta public:
        :param layer: A fully-connected layer
        :return: Number of accumulate operations for addressing a fully-connected layer in a formal neural network
        """  # noqa: RUF002
        return layer.input_shape[0][-1] * layer.output_shape[0][-1]

    def _e_addr_fc_fnn(self, layer: TDenseLayer) -> float:
        """Compute energy for addressing a fully-connected layer in a formal neural network.

        :meta public:
        :param layer: A fully-connected layer
        :return: Energy for addressing a fully-connected layer in a formal neural network.
        """
        return self._mac_addr_fc_fnn(layer) * (self._e_mul + self._e_add) + self._acc_addr_fc_fnn(layer) * self._e_add

    def _compute_model_energy_fnn(self,
                                  modelgraph: ModelGraph,
                                  e_rdram: Callable[[int], float],
                                  e_wrram: Callable[[int], float]) -> list[EnergyMetrics]:
        """Compute the energy per inference for each layer of a formal neural network.

        Supports the following layers:

        * :class:`qualia_codegen_core.graph.layers.TConvLayer.TConvLayer`
        * :class:`qualia_codegen_core.graph.layers.TDenseLayer.TDenseLayer`

        :meta public:
        :param modelgraph: Model to computer energy on
        :param e_rdram: Function to compute memory read energy for a given memory size
        :param e_wrram: Function to computer memory write energy for a given memory size
        :return: A list of EnergyMetrics for each layer and a total with fields populated with energy estimation
        """
        from qualia_codegen_core.graph.layers import TConvLayer, TDenseLayer

        ems: list[EnergyMetrics] = []
        for node in modelgraph.nodes:
            if isinstance(node.layer, TConvLayer):
                em = EnergyMetrics(name=node.layer.name,
                                   mem_pot=0,
                                   mem_weights=self._e_rdweights_conv_fnn(node.layer, e_rdram),
                                   mem_bias=self._e_rdbias_conv_fnn(node.layer, e_rdram),
                                   mem_io=self._e_rdin_conv_fnn(node.layer, e_rdram) + self._e_wrout_conv_fnn(node.layer, e_wrram),
                                   ops=self._e_ops_conv_fnn(node.layer),
                                   addr=self._e_addr_conv_fnn(node.layer),
                                   input_spikerate=None,
                                   output_spikerate=None,
                                   input_is_binary=False,
                                   output_is_binary=False)
            elif isinstance(node.layer, TDenseLayer):
                em = EnergyMetrics(name=node.layer.name,
                                   mem_pot=0,
                                   mem_weights=self._e_rdweights_fc_fnn(node.layer, e_rdram),
                                   mem_bias=self._e_rdbias_fc_fnn(node.layer, e_rdram),
                                   mem_io=self._e_rdin_fc_fnn(node.layer, e_rdram) + self._e_wrout_fc_fnn(node.layer, e_wrram),
                                   ops=self._e_ops_fc_fnn(node.layer),
                                   addr=self._e_addr_fc_fnn(node.layer),
                                   input_spikerate=None,
                                   output_spikerate=None,
                                   input_is_binary=False,
                                   output_is_binary=False)
            else:
                logger.warning('%s skipped, result may be inaccurate', node.layer.name)
                continue

            ems.append(em)

        em_total = EnergyMetrics(name='Total',
                                 mem_pot=sum(em.mem_pot for em in ems),
                                 mem_weights=sum(em.mem_weights for em in ems),
                                 mem_bias=sum(em.mem_bias for em in ems),
                                 mem_io=sum(em.mem_io for em in ems),
                                 ops=sum(em.ops for em in ems),
                                 addr=sum(em.addr for em in ems),
                                 input_spikerate=None,
                                 output_spikerate=None,
                                 input_is_binary=False,
                                 output_is_binary=False)
        ems.append(em_total)

        return ems


    #######
    # SNN #
    #######

    def _rdin_snn(self, layer: TBaseLayer, input_spikerate: float) -> float:
        """Count average number of read operations for the input of a layer in a spiking neural network.

        Eq. 5: θl-1.

        :meta public:
        :param layer: A convolutional or fully-connected layer
        :param input_spikerate: Average spike per input per inference
        :return: Average number of read operations for the input of a layer in a spiking neural network
        """
        theta_in = input_spikerate * math.prod(layer.input_shape[0][1:])
        return theta_in  # noqa: RET504

    def _e_rdin_snn(self, layer: TBaseLayer, input_spikerate: float, e_rdram: Callable[[int], float]) -> float:
        """Compute average energy for read operations for the input of a layer in a spiking neural network.

        :meta public:
        :param layer: A convolutional or fully-connected layer
        :param input_spikerate: Average spike per input per inference
        :param e_rdram: Function to compute memory read access energy for a given memory size
        :return: Average energy for read operations for the input of a layer in a spiking neural network
        """
        return self._rdin_snn(layer, input_spikerate) * e_rdram(self._fifo_size)


    def _rdweights_conv_snn(self, layer: TConvLayer, input_spikerate: float) -> float:
        """Count average number of read operations for weights of convolutional layer (excluding bias) in spiking neural network.

        Eq. 7f: θl-1 × Cout × Wkernel × Hkernel.

        :meta public:
        :param layer: A convolutional layer
        :param input_spikerate: Average spike per input per inference
        :return: Average number of read operations for weights of convolutional layer (excluding bias) in spiking neural network
        """  # noqa: RUF002
        theta_in = input_spikerate * math.prod(layer.input_shape[0][1:])
        return  theta_in * layer.output_shape[0][-1] * math.prod(layer.kernel_size)

    def _e_rdweights_conv_snn(self, layer: TConvLayer, input_spikerate: float, e_rdram: Callable[[int], float]) -> float:
        """Compute average energy for read operations for convolutional layer weights (excluding bias) in spiking neural network.

        :meta public:
        :param layer: A convolutional layer
        :param input_spikerate: Average spike per input per inference
        :param e_rdram: Function to compute memory read access energy for a given memory size
        :return: Average energy for read operations for weights of convolutional layer (excluding bias) in spiking neural network
        """
        return self._rdweights_conv_snn(layer, input_spikerate) * e_rdram(layer.kernel.size)

    def _rdbias_conv_snn(self, layer: TConvLayer) -> int:
        """Count number of read operations for the biases of a convolutional layer in a spiking neural network.

        Eq. 7s: Cout × Wout × Hout.

        :meta public:
        :param layer: A convolutional layer
        :return: Number of read operations for the biases of a convolutional layer in a spiking neural network
        """  # noqa: RUF002
        return math.prod(layer.output_shape[0][1:])

    def _e_rdbias_conv_snn(self, layer: TConvLayer, e_rdram: Callable[[int], float]) -> float:
        """Compute energy for read operations for the biases of a convolutional layer in a spiking neural network.

        :meta public:
        :param layer: A convolutional layer
        :param e_rdram: Function to compute memory read access energy for a given memory size
        :return: Energy for read operations for the biases of a convolutional layer in a spiking neural network or 0 if the layer
                 does not use biases
        """
        if layer.use_bias:
            return self._rdbias_conv_snn(layer) * e_rdram(layer.bias.size)
        return 0


    def _rdweights_fc_snn(self, layer: TDenseLayer, input_spikerate: float) -> float:
        """Count average number of read operations for weights of fully-connected layer (excluding bias) in spiking neural network.

        Eq. 8f: θl-1 × Nout.

        :meta public:
        :param layer: A fully-connected layer
        :param input_spikerate: Average spike per input per inference
        :return: Average number of read operations for weights of fully-connected layer (excluding bias) in spiking neural network
        """  # noqa: RUF002
        theta_in = input_spikerate * math.prod(layer.input_shape[0][1:])
        return theta_in * layer.output_shape[0][-1]

    def _e_rdweights_fc_snn(self, layer: TDenseLayer, input_spikerate: float, e_rdram: Callable[[int], float]) -> float:
        """Compute average energy for weights read operations of fully-connected layer (excluding bias) in spiking neural network.

        :meta public:
        :param layer: A fully-connected layer
        :param input_spikerate: Average spike per input per inference
        :param e_rdram: Function to compute memory read access energy for a given memory size
        :return: Average energy for read operations for weights of fully-connected layer (excluding bias) in spiking neural network
        """
        return self._rdweights_fc_snn(layer, input_spikerate) * e_rdram(layer.kernel.size)

    def _rdbias_fc_snn(self, layer: TDenseLayer) -> int:
        """Count number of read operations for the biases of a fully-connected layer in a spiking neural network.

        Eq. 6.2s: Nout.

        :meta public:
        :param layer: A fully-connected layer
        :return: Number of read operations for the biases of a fully-connected layer in a spiking neural network
        """
        return layer.output_shape[0][-1]

    def _e_rdbias_fc_snn(self, layer: TDenseLayer, e_rdram: Callable[[int], float]) -> float:
        """Compute energy for read operations for the biases of a fully-connected layer in a spiking neural network.

        :meta public:
        :param layer: A fully-connected layer
        :param e_rdram: Function to compute memory read access energy for a given memory size
        :return: Energy for read operations for the biases of a fully-connected layer in a spiking neural network or 0 if the layer
            does not use biases
        """
        if layer.use_bias:
            return self._rdbias_fc_snn(layer) * e_rdram(layer.bias.size)
        return 0


    def _wrout_snn(self, layer: TBaseLayer, output_spikerate: float) -> float:
        """Count average number of write operations for the output of a layer in a spiking neural network.

        Eq. 13: Noutput.

        :meta public:
        :param layer: A convolutional or fully-connected layer
        :param output_spikerate: Average spike per output per inference
        :return: Average number of write operations for the output of a layer in a spiking neural network
        """
        theta_out = output_spikerate * math.prod(layer.output_shape[0][1:])
        return theta_out  # noqa: RET504

    def _e_wrout_snn(self, layer: TBaseLayer, output_spikerate: float, e_wrram: Callable[[int], float]) -> float:
        """Compute average energy for write operations for the output of a layer in a spiking neural network.

        :meta public:
        :param layer: A convolutional or fully-connected layer
        :param output_spikerate: Average spike per output per inference
        :param e_rdram: Function to compute memory write access energy for a given memory size
        :return: Average energy for write operations for the output of a layer in a spiking neural network
        """
        return self._wrout_snn(layer, output_spikerate) * e_wrram(self._fifo_size)


    def _wrpot_conv_snn(self, layer: TConvLayer, input_spikerate: float) -> float:
        """Count average number of write operations for the potentials of a convolutional layer in a spiking neural network.

        Eq 14: θl-1 × Cout × Wkernel × Hkernel + Cout × Hout × Wout.

        :meta public:
        :param layer: A convolutional layer
        :param input_spikerate: Average spike per input per inference
        :return: Average number of write operations for the potentials of a convolutional layer in a spiking neural network
        """  # noqa: RUF002
        theta_in = input_spikerate * math.prod(layer.input_shape[0][1:])
        return theta_in * layer.output_shape[0][-1] * math.prod(layer.kernel_size) + math.prod(layer.output_shape[0][1:])

    def _e_wrpot_conv_snn(self, layer: TConvLayer, input_spikerate: float, e_wrram: Callable[[int], float]) -> float:
        """Compute average energy for write operations for the potentials of a convolutional layer in a spiking neural network.

        :meta public:
        :param layer: A convolutional layer
        :param input_spikerate: Average spike per input per inference
        :param e_rdram: Function to compute memory write access energy for a given memory size
        :return: Average energy for write operations for the potentials of a convolutional layer in a spiking neural network
        """
        return self._wrpot_conv_snn(layer, input_spikerate) * e_wrram(math.prod(layer.output_shape[0][1:]))

    def _wrpot_fc_snn(self, layer: TDenseLayer, input_spikerate: float) -> float:
        """Count average number of write operations for the potentials of a fully-connected layer in a spiking neural network.

        Eq 15: θl-1 × Nout + Nout.

        :meta public:
        :param layer: A fully-connected layer
        :param input_spikerate: Average spike per input per inference
        :return: Average number of write operations for the potentials of a fully-connected layer in a spiking neural network
        """  # noqa: RUF002
        theta_in = input_spikerate * math.prod(layer.input_shape[0][1:])
        return theta_in * layer.output_shape[0][-1] + layer.output_shape[0][-1]

    def _e_wrpot_fc_snn(self, layer: TDenseLayer, input_spikerate: float, e_wrram: Callable[[int], float]) -> float:
        """Compute average energy for write operations for the potentials of a fully-connected layer in a spiking neural network.

        :meta public:
        :param layer: A fully-connected layer
        :param input_spikerate: Average spike per input per inference
        :param e_rdram: Function to compute memory write access energy for a given memory size
        :return: Average energy for write operations for the potentials of a fully-connected layer in a spiking neural network
        """
        return self._wrpot_fc_snn(layer, input_spikerate) * e_wrram(math.prod(layer.output_shape[0][1:]))


    def _mac_ops_conv_snn(self, layer: TConvLayer, timesteps: int, leak: bool) -> int:  # noqa: FBT001
        """Count number of multiply-accumulate operations inside a convolutional layer in a spiking neural network.

        Eq. 1.3: T × Cout × Hout × Wout.

        :meta public:
        :param layer: A convolutional layer
        :param timesteps: Number of timesteps
        :param leak: Whether the neuron has leak (LIF) or not (IF)
        :return: Number of multiply-accumulate operations inside a convolutional layer in a spiking neural network, 0 if no leak
        """  # noqa: RUF002
        if not leak:
            return 0
        return timesteps * math.prod(layer.output_shape[0][1:])

    def _acc_ops_conv_snn(self, layer: TConvLayer, input_spikerate: float, output_spikerate: float, timesteps: int) -> float:
        """Count average number of accumulate operations inside a convolutional layer in a spiking neural network.

        Eq. 1.4: θl-1 × ceil(Hkernel / S) × ceil(Wkernel / S) × Cout + T × Cout × Hout × Wout + θl.

        :meta public:
        :param layer: A convolutional layer
        :param input_spikerate: Average spike per input per inference
        :param output_spikerate: Average spike per output per inference
        :param timesteps: Number of timesteps
        :return: Average number of accumulate operations inside a convolutional layer in a spiking neural network.
        """  # noqa: RUF002
        theta_in = input_spikerate * math.prod(layer.input_shape[0][1:])
        theta_out = output_spikerate * math.prod(layer.output_shape[0][1:])
        return (theta_in
                * math.prod(math.ceil(k / s) for k, s in zip(layer.kernel_size, layer.strides))
                * layer.output_shape[0][-1]
                + timesteps * math.prod(layer.output_shape[0][1:])
                + theta_out)

    def _e_ops_conv_snn(self,  # noqa: PLR0913
                      layer: TConvLayer,
                      input_spikerate: float,
                      output_spikerate: float,
                      timesteps: int,
                      leak: bool) -> float:  # noqa: FBT001
        """Compute average energy for multiply-accumulate and accumulate inside a convolutional layer in spiking neural network.

        :meta public:
        :param layer: A convolutional layer
        :param input_spikerate: Average spike per input per inference
        :param output_spikerate: Average spike per output per inference
        :param timesteps: Number of timesteps
        :param leak: Whether the neuron has leak (LIF) or not (IF)
        :return: Average energy for multiply-accumulate and accumulate inside a convolutional layer in spiking neural network.
        """
        return (self._mac_ops_conv_snn(layer, timesteps, leak) * (self._e_mul + self._e_add)
                + self._acc_ops_conv_snn(layer, input_spikerate, output_spikerate, timesteps) * self._e_add)


    def _mac_ops_fc_snn(self, layer: TDenseLayer, timesteps: int, leak: bool) -> int:  # noqa: FBT001
        """Count number of multiply-accumulate operations inside a fully-connected layer in a spiking neural network.

        Eq. 2.3: T × Nout.

        :meta public:
        :param layer: A fully-connected layer
        :param timesteps: Number of timesteps
        :param leak: Whether the neuron has leak (LIF) or not (IF)
        :return: Number of multiply-accumulate operations inside a fully-connected layer in a spiking neural network, 0 if no leak
        """  # noqa: RUF002
        if not leak:
            return 0
        return timesteps * layer.output_shape[0][-1]

    def _acc_ops_fc_snn(self, layer: TDenseLayer, input_spikerate: float, output_spikerate: float, timesteps: int) -> float:
        """Count average number of accumulate operations inside a fully-connected layer in a spiking neural network.

        Eq. 2.4: θl-1 × Nout + T × Nout + θl.

        **Warning 1**: the article counts Nin times too many as 1 input spike triggers only Nout MAC (each output neuron).

        **Warning 2**: the article is missing the third term which corresponds to the reset.

        Original equation (not used here):
        Eq. 2.4: θl-1 × Nin × Nout + T × Nout.

        :meta public:
        :param layer: A fully-connected layer
        :param input_spikerate: Average spike per input per inference
        :param output_spikerate: Average spike per output per inference
        :param timesteps: Number of timesteps
        :return: Average number of accumulate operations inside a fully-connected layer in a spiking neural network.
        """  # noqa: RUF002
        theta_in = input_spikerate * math.prod(layer.input_shape[0][1:])
        return (theta_in * layer.output_shape[0][-1]
                + timesteps * layer.output_shape[0][-1]
                + output_spikerate * math.prod(layer.output_shape[0][1:]))

    def _e_ops_fc_snn(self,  # noqa: PLR0913
                      layer: TDenseLayer,
                      input_spikerate: float,
                      output_spikerate: float,
                      timesteps: int,
                      leak: bool) -> float:  # noqa: FBT001
        """Compute average energy for multiply-accumulate and accumulate inside fully-connected layer in spiking neural network.

        :meta public:
        :param layer: A fully-connected layer
        :param input_spikerate: Average spike per input per inference
        :param output_spikerate: Average spike per output per inference
        :param timesteps: Number of timesteps
        :param leak: Whether the neuron has leak (LIF) or not (IF)
        :return: Average energy for multiply-accumulate and accumulate inside a fully-connected layer in spiking neural network.
        """
        return (self._mac_ops_fc_snn(layer, timesteps, leak) * (self._e_mul + self._e_add)
                + self._acc_ops_fc_snn(layer, input_spikerate, output_spikerate, timesteps) * self._e_add)


    def _mac_addr_conv_snn(self, layer: TConvLayer, input_spikerate: float) -> float:
        """Count average number of multiply-accumulate operations for addressing a convolutional layer in a spiking neural network.

        Eq. 16.2: θl-1 × 2.

        :meta public:
        :param layer: A convolutional layer
        :param input_spikerate: Average spike per input per inference
        :return: average number of multiply-accumulate operations for addressing a convolutional layer in a spiking neural network.
        """  # noqa: RUF002
        theta_in = input_spikerate * math.prod(layer.input_shape[0][1:])
        return theta_in * 2

    def _acc_addr_conv_snn(self, layer: TConvLayer, input_spikerate: float) -> float:
        """Count average number of accumulate operations for addressing a convolutional layer in a spiking neural network.

        Eq 16.3: θl-1 × Cout × Hkernel × Wkernel.

        :meta public:
        :param layer: A convolutional layer
        :param input_spikerate: Average spike per input per inference
        :return: Average number of accumulate operations for addressing a convolutional layer in a spiking neural network
        """  # noqa: RUF002
        theta_in = input_spikerate * math.prod(layer.input_shape[0][1:])
        return theta_in * math.prod(layer.kernel_size)

    def _e_addr_conv_snn(self, layer: TConvLayer, input_spikerate: float) -> float:
        """Compute average energy for addressing a convolutional layer in a spiking neural network.

        :meta public:
        :param layer: A convolutional layer
        :param input_spikerate: Average spike per input per inference
        :return: Average energy for addressing a convolutional layer in a spiking neural network.
        """
        return (self._mac_addr_conv_snn(layer, input_spikerate) * (self._e_mul + self._e_add)
                + self._acc_addr_conv_snn(layer, input_spikerate) * self._e_add)


    def _mac_addr_fc_snn(self, _: TDenseLayer) -> int:
        """Count number of multiply-accumulate operations for addressing a fully-connected layer in a spiking neural network.

        0

        :meta public:
        :return: 0
        """
        return 0

    def _acc_addr_fc_snn(self, layer: TDenseLayer, input_spikerate: float) -> float:
        """Count average number of accumulate operations for addressing a fully-connected layer in a spiking neural network.

        Eq. 17.2: θl-1 × Nout.

        :meta public:
        :param layer: A fully-connected layer
        :param input_spikerate: Average spike per input per inference
        :return: Average number of accumulate operations for addressing a fully-connected layer in a spiking neural network
        """  # noqa: RUF002
        return input_spikerate * layer.output_shape[0][-1]

    def _e_addr_fc_snn(self, layer: TDenseLayer, input_spikerate: float) -> float:
        """Compute average energy for addressing a fully-connected layer in a spiking neural network.

        :meta public:
        :param layer: A fully-connected layer
        :param input_spikerate: Average spike per input per inference
        :return: Average energy for addressing a fully-connected layer in a spiking neural network.
        """
        return (self._mac_addr_fc_snn(layer) * (self._e_mul + self._e_add)
                + self._acc_addr_fc_snn(layer, input_spikerate) * self._e_add)



    def _compute_model_energy_snn(self,  # noqa: PLR0913
                                  modelgraph: ModelGraph,
                                  input_spikerates: dict[str, float],
                                  output_spikerates: dict[str, float],
                                  input_is_binary: dict[str, bool],
                                  output_is_binary: dict[str, bool],
                                  timesteps: int,
                                  e_rdram: Callable[[int], float],
                                  e_wrram: Callable[[int], float]) -> list[EnergyMetrics]:
        """Compute the energy per inference for each layer of a spiking neural network.

        Supports the following layers:

        * :class:`qualia_codegen_core.graph.layers.TConvLayer.TConvLayer`
        * :class:`qualia_codegen_core.graph.layers.TDenseLayer.TDenseLayer`

        Input spike rates are per-timestep, this function multiplies by the number of timesteps to get the spike rates per infernce
        which are used by the operation count and energy computation functions.

        :meta public:
        :param modelgraph: Model to computer energy on
        :param input_spikerate: Dict of layer names and average spike per input per timestep for the layer
        :param output_spikerate: Dict of layer names and average spike per output per timestep for the layer
        :param input_is_binary: Dict of layer names and whether its input is binary (spike) or not
        :param output_is_binary: Dict of layer names and whether its output is binary (spike) or not
        :param timesteps: Number of timesteps
        :param e_rdram: Function to compute memory read energy for a given memory size
        :param e_wrram: Function to computer memory write energy for a given memory size
        :return: A list of EnergyMetrics for each layer and a total with fields populated with energy estimation
        """
        from qualia_codegen_core.graph.layers import TConvLayer, TDenseLayer
        from qualia_codegen_plugin_snn.graph.layers import TLifLayer

        ems: list[EnergyMetrics] = []
        for node in modelgraph.nodes:
            if not isinstance(node.layer, (TConvLayer, TDenseLayer)):
                logger.warning('%s skipped, result may be inaccurate', node.layer.name)
                continue

            # Account for timesteps here since the spikerate has been averaged over timesteps
            input_spikerate = input_spikerates[node.layer.name] * timesteps
            output_spikerate = output_spikerates[node.layer.name] * timesteps

            leak = len(node.outnodes) > 0 and isinstance(node.outnodes[0].layer, TLifLayer)

            if isinstance(node.layer, TConvLayer):
                em = EnergyMetrics(name=node.layer.name,
                                   mem_pot=self._e_wrpot_conv_snn(node.layer, input_spikerate, e_wrram),
                                   mem_weights=self._e_rdweights_conv_snn(node.layer, input_spikerate, e_rdram),
                                   mem_bias=self._e_rdbias_conv_snn(node.layer, e_rdram),
                                   mem_io=self._e_rdin_snn(node.layer, input_spikerate, e_rdram)
                                   + self._e_wrout_snn(node.layer, output_spikerate, e_wrram),
                                   ops=self._e_ops_conv_snn(node.layer, input_spikerate, output_spikerate, timesteps, leak),
                                   addr=self._e_addr_conv_snn(node.layer, input_spikerate),
                                   input_spikerate=input_spikerates[node.layer.name],
                                   output_spikerate=output_spikerates[node.layer.name],
                                   input_is_binary=input_is_binary[node.layer.name],
                                   output_is_binary=output_is_binary[node.layer.name],
                                   )
                ems.append(em)
            elif isinstance(node.layer, TDenseLayer):
                em = EnergyMetrics(name=node.layer.name,
                                   mem_pot=self._e_wrpot_fc_snn(node.layer, input_spikerate, e_wrram),
                                   mem_weights=self._e_rdweights_fc_snn(node.layer, input_spikerate, e_rdram),
                                   mem_bias=self._e_rdbias_fc_snn(node.layer, e_rdram),
                                   mem_io=self._e_rdin_snn(node.layer, input_spikerate, e_rdram)
                                   + self._e_wrout_snn(node.layer, output_spikerate, e_wrram),
                                   ops=self._e_ops_fc_snn(node.layer, input_spikerate, output_spikerate, timesteps, leak),
                                   addr=self._e_addr_fc_snn(node.layer, input_spikerate),
                                   input_spikerate=input_spikerates[node.layer.name],
                                   output_spikerate=output_spikerates[node.layer.name],
                                   input_is_binary=input_is_binary[node.layer.name],
                                   output_is_binary=output_is_binary[node.layer.name],
                                   )
                ems.append(em)

        em_total = EnergyMetrics(name='Total',
                                 mem_pot=sum(em.mem_pot for em in ems),
                                 mem_weights=sum(em.mem_weights for em in ems),
                                 mem_bias=sum(em.mem_bias for em in ems),
                                 mem_io=sum(em.mem_io for em in ems),
                                 ops=sum(em.ops for em in ems),
                                 addr=sum(em.addr for em in ems),
                                 input_spikerate=input_spikerates['__TOTAL__'],
                                 output_spikerate=output_spikerates['__TOTAL__'],
                                 input_is_binary=input_is_binary[modelgraph.nodes[1].layer.name], # First layer after 'input'
                                 output_is_binary=output_is_binary[modelgraph.nodes[-1].layer.name],
                                 )
        ems.append(em_total)

        return ems


    def _energy_summary(self, ems: list[EnergyMetrics]) -> str:
        """Generate a human-friendly text summary of the energy metrics per layer.

        :meta public:
        :param ems: List of EnergyMetrics per layer and the total
        :return: The text summary
        """
        pad = 12

        header = f'{"Layer": <{pad}} | {"EMemPot": <{pad}} | {"EMemWeights": <{pad}} | {"EMemBias": <{pad}} | {"EMemIO": <{pad}} |'
        header += f' {"EMemTotal": <{pad}} | {"EOps": <{pad}} | {"EAddr": <{pad}} | {"EOpsAddr": <{pad}} | {"ETotal": <{pad}} |'
        header += f' {"Input Spike Rate": <{pad}} | {"Output Spike Rate": <{pad}}\n'
        s = '—' * len(header) + '\n'
        s += header
        s += '—' * len(header) + '\n'
        for i, em in enumerate(ems):
            # Print in nJ, original values are in pJ
            s += f'{em.name: <{pad}.{pad}} | {em.mem_pot/1000: <{pad}.4} | {em.mem_weights/1000: <{pad}.4} |'
            s += f' {em.mem_bias/1000: <{pad}.4} | {em.mem_io/1000: <{pad}.4} | {em.mem_total/1000: <{pad}.4} |'
            s += f' {em.ops/1000: <{pad}.4} | {em.addr/1000: <{pad}.4} |'
            s += f' {em.opsaddr/1000: <{pad}.4} | {em.total/1000: <{pad}.4} |'

            input_spikerate_pad = 16 if em.input_is_binary else 11
            if em.input_spikerate is not None:
                s += f' {em.input_spikerate: <{input_spikerate_pad}.4}'
            else:
                s += f' {"N/A": <{input_spikerate_pad}.16}'
            if not em.input_is_binary:
                s += ' (NB)'
            s += ' |'

            output_spikerate_pad = 16 if em.output_is_binary else 11
            if em.output_spikerate is not None:
                s += f' {em.output_spikerate: <{output_spikerate_pad}.4}'
            else:
                s += f' {"N/A": <{output_spikerate_pad}.16}'
            if not em.output_is_binary:
                s += ' (NB)'
            s += '\n'

            s += ('-' if i < len(ems) - 2 else '—') * len(header) + '\n'

        s += ' NB = Non binary data'

        return s

    def _compute_spikerate(self,  # noqa: PLR0913
                           trainresult: TrainResult,
                           framework: SpikingJelly,
                           model: SNN,
                           dataset: RawData,
                           total_exclude_nonbinary: bool = True) -> tuple[dict[str, float],  # noqa: FBT002, FBT001
                                                                          dict[str, float],
                                                                          dict[str, bool],
                                                                          dict[str, bool]]:
        """Compute spike rate averaged over each neuron and timestep.

        A forward hook is added to each layer in order to record its input and output during inference performed over the given
        dataset.

        The layer name is computed using the internal :mod:`torch.fx` functions :meth:`torch.fx.graph._Namespace.create_name`
        and :meth:`torch.fx.graph._snake_case` so that it is compatible with
        :class:`qualia_codegen_core.graph.ModelGraph.ModelGraph`.

        :meta public:
        :param trainresult: TrainResult containing configuration for :class:`pytorch_lightning.Trainer` such as
                            :attr:`microai_core.microai.TrainResult.batch_size`
                            and :attr:`microai_core.microai.TrainResult.dataaugmentations`
        :param framework: A :class:`microai_plugin_snn.learningframework.SpikingJelly.SpikingJelly` instance
        :param model: A model to evaluate spike rate on
        :param dataset: The data to use for inference
        :return: A tuple of 4 dicts, first 2 dicts map a layer name to its input and output spike rates per timestep respectively,
                 the other 2 dicts map a layer name to ``True`` if the input or outputs are binary respectively,
                 ``False`` otherwise.
        """
        import torch
        from torch.fx.graph import _Namespace, _snake_case
        from torch.nn import Module
        namespace = _Namespace()  # type: ignore[no-untyped-call]

        @dataclass
        class SpikeCounter:
            spike_count: Number
            size: int
            binary: bool

        if_inputs_spike_count_and_size: dict[str, SpikeCounter]  = {}
        if_outputs_spike_count_and_size: dict[str, SpikeCounter] = {}

        def is_binary(a: torch.Tensor) -> bool:
            return torch.equal(a, (a > 0).float())

        def hook(layername: str, _: nn.Module, x: torch.Tensor, output: torch.Tensor) -> None:
            # WARNING: single input tensor only
            if isinstance(x, tuple) and len(x) > 1:
                logger.error('%s: Multiple inputs not supported.', layername)

            input_0 = (x[0] if isinstance(x, tuple) else x)
            inputnp = input_0.count_nonzero().item()
            outputnp = output.count_nonzero().item()

            if layername not in if_inputs_spike_count_and_size:
                if_inputs_spike_count_and_size[layername] = SpikeCounter(spike_count=inputnp,
                                                                         size=input_0.numel(),
                                                                         binary=is_binary(input_0))
            else:
                if_inputs_spike_count_and_size[layername].spike_count += inputnp
                if_inputs_spike_count_and_size[layername].size += input_0.numel()
                if_inputs_spike_count_and_size[layername].binary &= is_binary(input_0)

            if layername not in if_outputs_spike_count_and_size:
                if_outputs_spike_count_and_size[layername] = SpikeCounter(spike_count=outputnp,
                                                                          size=output.numel(),
                                                                          binary=is_binary(output))
            else:
                if_outputs_spike_count_and_size[layername].spike_count += outputnp
                if_outputs_spike_count_and_size[layername].size += output.numel()
                if_outputs_spike_count_and_size[layername].binary &= is_binary(output)

        handles = [layer.register_forward_hook(functools.partial(hook,
                                                                 namespace.create_name(_snake_case(layername), layer)))
                   for layername, layer in cast(Generator[tuple[str, Module], None, None], model.named_modules())
                   if not isinstance(layer, Quantizer)] # No hook to register for Quantizer module

        # Implement framework.predict()
        _ = framework.predict(model=model,
                              dataset=dataset,
                              batch_size=trainresult.batch_size,
                              dataaugmentations=trainresult.dataaugmentations,
                              experimenttracking=None)

        for handle in handles:
            handle.remove()

        input_spikerates = {n: sc.spike_count / sc.size for n, sc in if_inputs_spike_count_and_size.items()}
        output_spikerates = {n: sc.spike_count / sc.size for n, sc in if_outputs_spike_count_and_size.items()}

        # Filter out non-binary inputs/outputs if total_exclude_binary is True
        if total_exclude_nonbinary:
            logger.warning('Non-binary inputs/outputs are excluded from the total spike rate computation.')
        else:
            logger.warning('Non-binary inputs/outputs are included in the total spike rate computation.')
        filetered_if_inputs_spike_count_and_size = {n: sc for n, sc in if_inputs_spike_count_and_size.items()
                                                        if not total_exclude_nonbinary or sc.binary}
        filetered_if_outputs_spike_count_and_size = {n: sc for n, sc in if_outputs_spike_count_and_size.items()
                                                        if not total_exclude_nonbinary or sc.binary}
        # Special value account for the total spike rate across the whole network
        input_spikerates['__TOTAL__'] = (sum(sc.spike_count for sc in filetered_if_inputs_spike_count_and_size.values())
                                         / sum(sc.size for sc in filetered_if_inputs_spike_count_and_size.values()))
        output_spikerates['__TOTAL__'] = (sum(sc.spike_count for sc in filetered_if_outputs_spike_count_and_size.values())
                                         / sum(sc.size for sc in filetered_if_outputs_spike_count_and_size.values()))

        input_is_binary = {n: sc.binary for n, sc in if_inputs_spike_count_and_size.items()}
        output_is_binary = {n: sc.binary for n, sc in if_outputs_spike_count_and_size.items()}

        return input_spikerates, output_spikerates, input_is_binary, output_is_binary

    @override
    def __call__(self,  # noqa: C901
                 trainresult: TrainResult,
                 model_conf: ModelConfigDict) -> tuple[TrainResult, ModelConfigDict]:
        """Compute energy estimation metric from Lemaire et al, 2022.

        Uses Qualia CodeGen in order to build a :class:`qualia_codegen_core.graph.ModelGraph.ModelGraph` that is easier to parse.
        Call either :meth:`_compute_model_energy_snn` or :meth:`_compute_model_energy_fnn` depending on whether the model is an
        SNN or an FNN.
        Print the resulting metrics and log them to a CSV file inside the `logs/<bench.name>/EnergyEstimationMetric` directory.

        :meta public:
        :param trainresult: TrainResult containing the SNN or FNN model, the dataset and the training configuration
        :param model_conf: Unused
        :return: The unmodified trainresult
        """
        import spikingjelly.activation_based.layer as sjl  # type: ignore[import-untyped]
        import spikingjelly.activation_based.neuron as sjn  # type: ignore[import-untyped]
        from qualia_codegen_core.graph.layers import TAddLayer
        from qualia_codegen_plugin_snn.graph import TorchModelGraph
        from qualia_codegen_plugin_snn.graph.layers import TIfLayer
        from torch import nn

        import qualia_plugin_snn.learningmodel.pytorch.layers.quantized_SNN_layers as qsjn
        import qualia_plugin_snn.learningmodel.pytorch.layers.spikingjelly.quantized_layers as qsjl
        import qualia_plugin_snn.learningmodel.pytorch.layers.spikingjelly.quantized_layers1d as qsjl1d
        import qualia_plugin_snn.learningmodel.pytorch.layers.spikingjelly.quantized_layers2d as qsjl2d

        model = trainresult.model

        # SpikingJelly wrapped layers
        custom_layers: dict[type[nn.Module],
                            Callable[[nn.Module, TBaseLayer], tuple[type[TBaseLayer], list[Any]]]] = {
                                    sjl.AvgPool1d: TorchModelGraph.MODULE_MAPPING[nn.AvgPool1d],
                                    sjl.AvgPool2d: TorchModelGraph.MODULE_MAPPING[nn.AvgPool2d],
                                    sjl.BatchNorm1d: TorchModelGraph.MODULE_MAPPING[nn.BatchNorm1d],
                                    sjl.BatchNorm2d: TorchModelGraph.MODULE_MAPPING[nn.BatchNorm2d],
                                    sjl.Conv1d: TorchModelGraph.MODULE_MAPPING[nn.Conv1d],
                                    sjl.Conv2d: TorchModelGraph.MODULE_MAPPING[nn.Conv2d],
                                    sjl.Dropout: TorchModelGraph.MODULE_MAPPING[nn.Dropout],
                                    sjl.Flatten: TorchModelGraph.MODULE_MAPPING[nn.Flatten],
                                    sjl.Linear: TorchModelGraph.MODULE_MAPPING[nn.Linear],
                                    sjl.MaxPool1d: TorchModelGraph.MODULE_MAPPING[nn.MaxPool1d],
                                    sjl.MaxPool2d: TorchModelGraph.MODULE_MAPPING[nn.MaxPool2d],
                                    Add:  lambda *_: (TAddLayer, []),

                                    QuantizedIdentity:TorchModelGraph.MODULE_MAPPING[nn.Identity],
                                    qsjl.QuantizedLinear: TorchModelGraph.MODULE_MAPPING[nn.Linear],
                                    qsjl1d.QuantizedBatchNorm1d: TorchModelGraph.MODULE_MAPPING[nn.BatchNorm1d],
                                    qsjl2d.QuantizedBatchNorm2d: TorchModelGraph.MODULE_MAPPING[nn.BatchNorm2d],
                                    qsjl1d.QuantizedConv1d: TorchModelGraph.MODULE_MAPPING[nn.Conv1d],
                                    qsjl2d.QuantizedConv2d: TorchModelGraph.MODULE_MAPPING[nn.Conv2d],
                                    qsjl1d.QuantizedMaxPool1d: TorchModelGraph.MODULE_MAPPING[nn.MaxPool1d],
                                    qsjl2d.QuantizedMaxPool2d: TorchModelGraph.MODULE_MAPPING[nn.MaxPool2d],
                                    qsjn.QuantizedIFNode: TorchModelGraph.MODULE_MAPPING[sjn.IFNode],
                                    qsjn.QuantizedLIFNode: TorchModelGraph.MODULE_MAPPING[sjn.LIFNode],
                                    QuantizedAdd:  lambda *_: (TAddLayer, []),
                                    }
        step_mode = getattr(model, 'step_mode', 's')
        if step_mode != 's':
            logger.error("Only single-step mode (step_mode='s') is supported, current step_mode='%s'", step_mode)
            return trainresult, model_conf

        modelgraph = TorchModelGraph(model).convert(custom_layers=custom_layers)
        if modelgraph is None:
            logger.error('Model graph conversion failed')
            return trainresult, model_conf

        def e_rdram(x: int) -> float:
            return self._e_ram(x, self._mem_width)
        def e_wrram(x: int) -> float:
            return self._e_ram(x, self._mem_width)
        logger.info('Memory width set to %s bits', self._mem_width)

        if getattr(model, 'is_snn', False) or isinstance(model, SNN):
            if not isinstance(trainresult.framework, SpikingJelly):
                logger.error('LearningFramework must be SpikingJelly or a derived class, got: %s', type(trainresult.framework))
                raise TypeError

            if not self._fifo_size:
                logger.error('fifo_size is required for SNN models')
                raise ValueError

            # Tuple unpacking
            (input_spikerates,
             output_spikerates,
             input_is_binary,
             output_is_binary) = self._compute_spikerate(trainresult,
                                                         trainresult.framework,
                                                         model,
                                                         trainresult.testset,
                                                         total_exclude_nonbinary=self._total_spikerate_exclude_nonbinary)

            # Copy spikerate from IF activation layer to previous layer since it is the one energy is computed on
            for node in modelgraph.nodes:
                if isinstance(node.layer, TIfLayer):
                    output_spikerates[node.innodes[0].layer.name] = output_spikerates[node.layer.name]
                    # Also copy binary output check
                    output_is_binary[node.innodes[0].layer.name] = output_is_binary[node.layer.name]

            ems = self._compute_model_energy_snn(modelgraph,
                                                 input_spikerates,
                                                 output_spikerates,
                                                 input_is_binary,
                                                 output_is_binary,
                                                 model.timesteps,
                                                 e_rdram,
                                                 e_wrram)
        else:
            ems = self._compute_model_energy_fnn(modelgraph,
                                                 e_rdram,
                                                 e_wrram)

        for em in ems:
            self.csvlogger(em.asnamedtuple())

        logger.info(('Estimated model energy consumption for one inference on 45nm ASIC (nJ)'
                    ' and spike rate per neuron per timestep:\n%s'),
                    self._energy_summary(ems))

        return trainresult, model_conf

    #def _e_ram(self, mem_params: int) -> float:
    #    """Linear interpolation with {(8192, 10), (32768, 20), (1048576, 100)} point set (B, pJ) using least-squares methods.
    #
    #    45nm SRAM"""
    #    0.000082816 * mem_params + 13.2563

    def _e_ram(self, mem_params: int, bits: int) -> float:
        """From Computation_cost_metric.ipynb 45nm SRAM.

        :meta public:
        :param mem_params: Number of element stored in this memory
        :param bits: Data width in bits
        :return: Energy for a single read or write access in this memory
        """
        return 1.09 * 10**(-5) * mem_params * bits + 13.2

    @property
    def _e_add(self) -> float:
        """Energy for a single add operation.

        :meta public:
        """
        return self._m_e_add

    @property
    def _e_mul(self) -> float:
        """Energy for a single mul operation.

        :meta public:
        """
        return self._m_e_mul



"""Provide a preprocessing module to split data into timesteps."""

import sys

from qualia_core.datamodel import RawDataModel
from qualia_core.preprocessing.Preprocessing import Preprocessing

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class Split2TimeSteps(Preprocessing[RawDataModel, RawDataModel]):
    """Preprocessing module to split 1D input dataset into multiple timesteps."""

    def __init__(self, chunks: int) -> None:
        """Construct :class:`qualia_plugin_snn.preprocessing.Split2TimeSteps.Split2TimeSteps`.

        :param chunks: Number of chunks to split the data into
        """
        super().__init__()

        self.__chunks = chunks

    @override
    def __call__(self, datamodel: RawDataModel) -> RawDataModel:
        """Split the given :class:`qualia_core.datamodel.RawDataModel.RawDataModel` into multiple timesteps.

        Input data should be 1D (+ channel) with [N, S, C] order (channels_last).
        Output data has [N, T, S // T, C] dimensions
        Extra data that do not fit in a chunk is truncated.

        :param datamodel: The input dataset
        :return: The dataset with additional timestep dimension
        """
        for _, s in datamodel:
            truncated_dim = (s.x.shape[1] // self.__chunks) * self.__chunks
            s.x = s.x[:,:truncated_dim,:]
            s.x = s.x.reshape((s.x.shape[0], self.__chunks, s.x.shape[1] // self.__chunks) + s.x.shape[2:])

        return datamodel



"""Qualia-Plugin-SNN postprocessing package contains postprocessing modules adapted for or dedicated to Spiking Neural Networks."""

from .EnergyEstimationMetric import EnergyEstimationMetric
from .FuseBatchNorm import FuseBatchNorm
from .QualiaCodeGen import QualiaCodeGen

__all__ = ['EnergyEstimationMetric', 'FuseBatchNorm', 'QualiaCodeGen']



"""Provide a postprocessing module to fuse BatchNorm layers to the previous convolution with SpikingJelly layers support."""

from __future__ import annotations

import sys

import spikingjelly.activation_based.base as sjb  # type: ignore[import-untyped]
import spikingjelly.activation_based.layer as sjl  # type: ignore[import-untyped]
import spikingjelly.activation_based.neuron as sjn  # type: ignore[import-untyped]
from qualia_core.postprocessing.FuseBatchNorm import FuseBatchNorm as FuseBatchNormQualiaCore
from qualia_core.typing import TYPE_CHECKING

# We are inside a TYPE_CHECKING block but our custom TYPE_CHECKING constant triggers TCH001-TCH003 so ignore them
if TYPE_CHECKING:
    from torch import nn  # noqa: TCH002
    from torch.fx.graph_module import GraphModule  # noqa: TCH002

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class FuseBatchNorm(FuseBatchNormQualiaCore):
    """Extend :class:`qualia_core.postprocessing.FuseBatchNorm.FuseBatchNorm` with support for Spiking Neural Networks.

    :class:`spikingjelly.activation_based.neuron.BaseNode` and :class:`spikingjelly.activation_based.base.StepModule`
    are added to :attr:`qualia_core.postprocessing.FuseBatchNorm.FuseBatchNorm.custom_layers` to avoid tracing inside.

    SpikingJelly-wrapped layers are added to lookup patterns:

    * (:class:`spikingjelly.activation_based.layer.Conv1d`, :class:`spikingjelly.activation_based.layer.BatchNorm1d`
    * (:class:`spikingjelly.activation_based.layer.Conv2d`, :class:`spikingjelly.activation_based.layer.BatchNorm2d`
    * (:class:`spikingjelly.activation_based.layer.Conv3d`, :class:`spikingjelly.activation_based.layer.BatchNorm3d`

    Extended attributes ``timesteps`` and ``is_snn`` are copied into target model.
    """

    custom_layers: tuple[type[nn.Module | sjb.StepModule], ...] = (
            sjn.BaseNode,
            sjb.StepModule, # StepModule not a subclass of nn.Module but still required to avoid parsing sj-wrapped nn layers
            *FuseBatchNormQualiaCore.custom_layers,
            )


    def __init__(self) -> None:
        """Construct :class:`qualia_plugin_snn.postprocessing.FuseBatchNorm.FuseBatchNorm`.

        Patterns are extended to include SpikingJelly-wrapped layers.
        """
        super().__init__()

        self.patterns += [
                (sjl.Conv1d, sjl.BatchNorm1d),
                (sjl.Conv2d, sjl.BatchNorm2d),
                (sjl.Conv3d, sjl.BatchNorm3d),
                ]

    @override
    def fuse(self,
             model: nn.Module,
             inplace: bool = False) -> GraphModule:
        """Fuse BatchNorm to Conv and copy source model ``timesteps`` and `is_snn` attributes to target model.

        :param mode: PyTorch model with Conv-BatchNorm layers to fuse
        :param inplace: Modify model in place instead of deep-copying
        :return: Resulting model with BatchNorm fused to Conv
        """
        fused_model = super().fuse(model, inplace=inplace)
        if hasattr(model, 'timesteps'):
            fused_model.timesteps = model.timesteps
        if hasattr(model, 'is_snn'):
            fused_model.is_snn = model.is_snn
        return fused_model



"""Provide the SpikingJelly single-step learningframework module."""

from __future__ import annotations

import sys

import torch
from qualia_core.learningframework import PyTorch
from qualia_core.typing import TYPE_CHECKING
from spikingjelly.activation_based import functional  # type: ignore[import-untyped]

import qualia_plugin_snn
import qualia_plugin_snn.learningmodel.pytorch

# We are inside a TYPE_CHECKING block but our custom TYPE_CHECKING constant triggers TCH001-TCH003 so ignore them
if TYPE_CHECKING:
    from types import ModuleType  # noqa: TCH003

    from qualia_plugin_snn.learningmodel.pytorch.SNN import SNN  # noqa: TCH001

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override


class SpikingJelly(PyTorch):
    """SpikingJelly single-step LearningFramework implementation extending PyTorch.

    :attr:`qualia_core.learningframework.PyTorch.PyTorch.learningmodels` are replaced by the Spiking Neural Networks
    from :mod:`qualia_plugin_snn.learningmodel.pytorch`
    """

    learningmodels: ModuleType = qualia_plugin_snn.learningmodel.pytorch
    """:mod:`qualia_plugin_snn.learningmodel.pytorch` additional learningmodels for Spiking Neural Networks.

    :meta hide-value:
    """

    class TrainerModule(PyTorch.TrainerModule):
        """SpikingJelly single-step TrainerModule implementation extending PyTorch TrainerModule."""

        #: Spiking Neural Network model used by this TrainerModule
        model: SNN

        @override
        def forward(self, x: torch.Tensor) -> torch.Tensor:
            """Forward pass for a Spiking Neural Network model with duplicated timesteps.

            First calls SpikingJelly's reset on the model to reset neurons potentials.
            Then duplicate the input to generate the number of timesteps given by
            :attr:`qualia_plugin_snn.learningmodel.pytorch.SNN.SNN.timesteps`
            Call :meth:`qualia_plugin_snn.learningmodel.pytorch.SNN.SNN.forward` for each timestep.
            Finally, average the output of the model over the timesteps.

            :param x: Input data
            :return: Output predictions
            """
            functional.reset_net(self.model)
            x_list = [self.model(x) for _ in range(self.model.timesteps)]

            return torch.stack(x_list).sum(0) / self.model.timesteps
